/*
*	Copyright(c) 2012 - 2022  IntervalZero, Inc.  All rights reserved.
*
*	Module Name:
*		RtIGCFuncs.c
*
*	Abstract:
*		This solutions shows the framework required for and RTX64 Network Interface Adapter(NIC)
*		driver.
*
*	Environments:
*		Real-time 64-bit Subsystem (supported configurations: RTSSDebug and RTSSRelease)
*
*	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
*	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
*	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
*	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
*	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*	POSSIBILITY OF SUCH DAMAGE.
*/

#include "RtNalIGC.h"

typedef  unsigned long volatile* pCriticalLock;

extern DRV_GLOBALS* pGlobals;

#define CALL_APP_LEVEL_CALLBACK			TRUE
#define DONT_CALL_APP_LEVEL_CALLBACK	FALSE

#if !((RTX64_MAJOR > 4) || (RTX64_MAJOR == 4 && RTX64_MINOR > 2))
#define RtndMediaSpeed2500	 5
#endif

static BOOL
ConfigureMsixMultiVectorLayout(DRVDEV* pCardInfo);

// Generate NAL monitor event
static void GenerateMonitorEvent(BYTE* macAddress, BYTE queueNumber, BYTE eventType, ULONG packetCnt, DWORD errorCode)
{
	BYTE buffer[RTNAL_MONITOR_DATA_SIZE];
	ULONG* pulBuffer;
	int i, n;
	static BOOL reported = FALSE;

	pulBuffer = (ULONG*)&buffer[RTNAL_MONITOR_DATA_MAJIC_OFFSET];
	*pulBuffer = RTNAL_MONITOR_EVENT_MAGIC;

	for (i = 0, n = RTNAL_MONITOR_DATA_MACADDR_OFFSET; i < 6; i++, n++)
		buffer[n] = macAddress[i];

	buffer[RTNAL_MONITOR_DATA_QUEUE_NUM_OFFSET] = queueNumber;
	buffer[RTNAL_MONITOR_DATA_EVENT_TYPE_OFFSET] = eventType;

	pulBuffer = (ULONG*)&buffer[RTNAL_MONITOR_DATA_PACKET_CNT_OFFSET];
	*pulBuffer = packetCnt;

	pulBuffer = (ULONG*)&buffer[RTNAL_MONITOR_DATA_ERROR_OFFSET];
	*pulBuffer = errorCode;

	if (!RtGenerateEvent(RTNAL_MONITOR_EVENT_KIND, buffer, sizeof buffer) && pGlobals->displayErrorMessages && !reported)
	{
		RtPrintf("Failed to generate NAL monitor event, error %d. Please increase monitoring memor\n", GetLastError());
		reported = TRUE;
	}

}

void RtIgcInitializeCriticalLock(HFREECRITICALLOCK* pLock, char* szIndentifier)
{
	strcpy(pLock->lockSymbolic, szIndentifier);
	pLock->lock = 0;
}

void RtIgcDeleteCriticalLock(HFREECRITICALLOCK* pLock)
{

}

void RtIgcEnterCriticalLock(HFREECRITICALLOCK* pLock)
{
	DWORD spins = 0;

	if (!InterlockedCompareExchange(&(pLock->lock), 1, 0))
	{
		// got the lock right away
		pLock->lockOwnerThreadID = GetCurrentThreadId();
		pLock->lockOwnerThreadPriority = RtGetThreadPriority(GetCurrentThread());
		return;
	}
	else
	{
		int Error = 0;

		HANDLE			hThread;
		int				newPriority, originalPriority;

		hThread = GetCurrentThread();
		newPriority = originalPriority = RtGetThreadPriority(hThread);
		pLock->lockRequestorThreadPriority = originalPriority;

		while (InterlockedCompareExchange(&(pLock->lock), 1, 0))
		{
			--newPriority;
			if (newPriority <= 0)
				newPriority = 1;
			RtSetThreadPriority(hThread, newPriority);

			spins++;

			if (spins > 10000)
			{	// Dead locked but we shouldn't be.  Developer has a problem in his design.
				// OK Give up.  We expect it is impossible  to reach this point.  But given murphey's law...
				// To help early development we return here with out actually having the lock.
				//  The print will show even if verbose mode is not set.
				RtSetThreadPriority(hThread, originalPriority);
				RtPrintf("RtIgcEnterCriticalLock. Excessive time in spin lock. Number of spins=%d\n", spins);
				RtPrintf("RtIgcEnterCriticalLock. Excessive time in spin lock. lockOwnerThreadID=%d\n", pLock->lockOwnerThreadID);
				RtPrintf("RtIgcEnterCriticalLock. Excessive time in spin lock. lockOwnerThreadPriority=%d\n", pLock->lockOwnerThreadPriority);
				RtPrintf("RtIgcEnterCriticalLock. Excessive time in spin lock. lockOwner Identifier=%s\n", pLock->lockSymbolic);
				RtPrintf("RtIgcEnterCriticalLock. Excessive time in spin lock. lockRequestor ThreadID=%d\n", GetCurrentThreadId());
				RtPrintf("RtIgcEnterCriticalLock. Excessive time in spin lock. lockRequestor starting ThreadPriority=%d\n", pLock->lockRequestorThreadPriority);
				RtPrintf("RtIgcEnterCriticalLock. Excessive time in spin lock. last number for max spins=%d\n", pLock->maxNumberOfSpins);
				return;
			}
			else if (spins > 127)
			{	// Dead locked and we are at low priority.
				// This can happen if the thread owns a mutex and is at another priority level
				// In that case RtSetThreadPriority won't call the scheduler
				// The last resort is to sleep for least amount of time.
				LARGE_INTEGER	Duration;

				RtSetThreadPriority(hThread, originalPriority);
				Duration.QuadPart = 1;  // Any nonzero value will sleep for 1 or 2 RtssHalClockTimerPeriods
				RtSleepFt(&Duration);
			}
		}

		// The while() loop stopped because we got the lock via the atomic operation
		// This is the expected path.  Fix up our priority and be done.
		RtSetThreadPriority(hThread, originalPriority);
	}
	// got the lock after some spins
	pLock->lockOwnerThreadID = GetCurrentThreadId();
	pLock->lockOwnerThreadPriority = RtGetThreadPriority(GetCurrentThread());
	if (spins > pLock->maxNumberOfSpins)
		pLock->maxNumberOfSpins = spins;
}


void RtIgcLeaveCriticalLock(HFREECRITICALLOCK* pLock)
{
	if (InterlockedCompareExchange(&(pLock->lock), 0, 1) != 1)
	{
		DebugMsg(("RtIgcEnterCriticalLock with invalid lock\n"));
		pLock->lock = 0;
	}
}

static BYTE* DrvPageAlignMemory(BYTE* pbyBlock, PHYSICAL_ADDRESS* pPysicalAddress)
{
	unsigned __int64 PageSize = SMALL_PAGE_SIZE;
	unsigned __int64 Offset = 0;
	unsigned __int64 Temp;

	*pPysicalAddress = RtGetPhysicalAddress(pbyBlock);

	Temp = (unsigned __int64)pPysicalAddress->QuadPart;
	Temp &= (PageSize - 1);
	if (Temp != 0)
	{
		Offset = (PageSize - Temp);
		pPysicalAddress->QuadPart += Offset;
	}
	return (pbyBlock + Offset);
}

BYTE* DrvAllocateMemory(BOOL fContiguous, DWORD dwSizeInBytes)
{
	BYTE* pvResult = NULL;

	if (fContiguous)
	{
		PHYSICAL_ADDRESS physaddrHighest;
		PHYSICAL_ADDRESS physaddrLowest;
		PHYSICAL_ADDRESS physaddrAlignmentBoundary;

		physaddrHighest.QuadPart = 0xFFFFFFFFFFFFFFFF;
		physaddrLowest.QuadPart = 0x0;
		physaddrAlignmentBoundary.QuadPart = 0x0;
		pvResult = (BYTE*)RtAllocateContiguousMemorySpecifyCache(dwSizeInBytes, physaddrLowest, physaddrHighest, physaddrAlignmentBoundary, MmCached);
	}
	else
	{
		pvResult = (BYTE*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, dwSizeInBytes);
	}
	if (NULL == pvResult)
	{
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);
	}

	return (pvResult);
}

void
DrvFreeMemory(BOOL fContiguous, PVOID pMem)
{
	if (fContiguous)
	{
		RtFreeContiguousMemory(pMem);
	}
	else
	{
		HeapFree(GetProcessHeap(), 0, pMem);
	}
}

/**--------------------------------------------------------------------------------------------------
   \fn	DRVDEV * DrvCreateDevice()

   \brief	DrvCreateDevice creates a local configuration structure array to hold configuration
			information read in during the RtndInitializeInterface function. The driver then simply
			increments a local instance count variable to index the array.   A copy of the device
			name and a pointer to a network device data structure are saved to the local
			configuration structure array slot along with other configuration information, so
			that RtndConfigure can find the proper slot in the local configuration array.

   \return	null if it fails, else a DRVDEV*.

*-----------------------------------------------------------------------------------------------**/
DRVDEV* DrvCreateDevice()
{
	pGlobals->drvCards[pGlobals->nCards] = (PDRVDEV)DrvAllocateMemory(TRUE, sizeof(struct _DRVDEV));

	if (!pGlobals->drvCards[pGlobals->nCards])
	{
		DebugMsg(("RtCreateDevice failed to allocate device\n"));
		return(NULL);
	}
	return(pGlobals->drvCards[pGlobals->nCards]);
}


/**--------------------------------------------------------------------------------------------------
\fn	static ULONG WINAPI DrvLineStatusThread(DRVDEV *pCardInfo)

\brief	Line status monitoring thread. This function is not called, but is started as a
thread from RtndConfigure.

\param [in,out]	pCardInfo	Pointer to the local configuration structure for this instance of
the driver.

\return	None.

*-----------------------------------------------------------------------------------------------**/
static ULONG WINAPI DrvLineStatusThread(DRVDEV* pCardInfo)
{
	// Current link status
	DWORD WaitStat = WAIT_FAILED;
	RTND_MEDIA_CONNECT_STATE linkStatus = RtndMediaConnectStateUnknown;
	RTND_MEDIA_SPEED currentLineSpeed = RtndMediaSpeedUnknown;
	// Flag that shows if link status was already updated 
	BOOL linkStatusWasUpdated = FALSE;

	RT_VERSION_INFO versionInfo;
	 
	if (!RtGetRuntimeVersionEx(&versionInfo))
	{
		 versionInfo.majorVersion = 0;
		 versionInfo.minorVersion = 0;
	}

	for (;;)
	{
		bool link_up;
		s32 retval;
		s32 duplex = 0;
		s32 speed = 0;

		WaitStat = RtWaitForSingleObject(pCardInfo->hLineStatusEvent, INFINITE);

		if (WaitStat != WAIT_OBJECT_0)
			break;

		RtResetEvent(pCardInfo->hLineStatusEvent);
		
		retval = RtCheckLink(pCardInfo, &speed, &link_up, &duplex);
		
		if (retval != IGC_SUCCESS)
		{
			pCardInfo->LinkIsActive = RtndMediaConnectStateUnknown;
			pCardInfo->LineSpeed = RtndMediaSpeedUnknown;
			pCardInfo->FullDuplex = RtndMediaDuplexStateUnknown;
		}
		else if (link_up)
		{
			pCardInfo->LinkIsActive = RtndMediaConnectStateConnected;
			if (duplex == FULL_DUPLEX)
				pCardInfo->FullDuplex = RtndMediaDuplexStateFull;
			else if (duplex == FULL_DUPLEX)
				pCardInfo->FullDuplex = RtndMediaDuplexStateHalf;
			else
				pCardInfo->FullDuplex = RtndMediaDuplexStateUnknown;


			if (speed == SPEED_2500)
			{
				 if ((versionInfo.majorVersion > 4) || (versionInfo.majorVersion == 4 && versionInfo.minorVersion > 2)) {
				pCardInfo->LineSpeed = RtndMediaSpeed2500;
			}
				 else {
					  pCardInfo->LineSpeed = RtndMediaSpeed10000;
				 }
			}
			else if (speed == SPEED_1000)
			{
				pCardInfo->LineSpeed = RtndMediaSpeed1000;
			}
			else if (speed == SPEED_100)
			{
				pCardInfo->LineSpeed = RtndMediaSpeed100;
			}
			else if (speed == SPEED_10)
			{
				pCardInfo->LineSpeed = RtndMediaSpeed10;
			}
			else
			{
				pCardInfo->LineSpeed = RtndMediaSpeedUnknown;
			}
		}
		else
		{
			pCardInfo->LinkIsActive = RtndMediaConnectStateDisConnected;
			pCardInfo->LineSpeed = RtndMediaSpeedUnknown;
			pCardInfo->FullDuplex = RtndMediaDuplexStateUnknown;
		}

		// Link status has not been reported yet or it changed
		if (FALSE == linkStatusWasUpdated || pCardInfo->LinkIsActive != linkStatus)
		{
			RTND_REQUEST RtndReq;
			RTND_STATUS  Status;
			DWORD bufData;

			linkStatus = pCardInfo->LinkIsActive;
#ifdef _DEBUG
			if (pCardInfo->LinkIsActive == RtndMediaConnectStateUnknown)
			{
				DebugMsg(("RtndMediaConnectStateUnknown\n"));
			}
			else if (pCardInfo->LinkIsActive == RtndMediaConnectStateConnected)
			{
				DebugMsg(("RtndMediaConnectStateConnected, LineSpeed %d\n", pCardInfo->LineSpeed));
			}
			else
			{
				DebugMsg(("RtndMediaConnectStateDisConnected\n"));
			}
#endif
			linkStatusWasUpdated = TRUE;

			// Signal all attached applications of change in link status
			RtndReq.InformationBuffer = &bufData;
			*((DWORD*)RtndReq.InformationBuffer) = pCardInfo->LinkIsActive;
			RtndReq.InformationBufferLength = sizeof bufData;
			RtndReq.Oid = RTND_OID_GEN_MEDIA_CONNECT_STATUS;

			RtnIndicateStatus(pCardInfo->ndp, &RtndReq, &Status);
		}
#ifdef TIMESTAMP_SUPPORT
		// Line speed changed. Update system timer increment.
		if ((pCardInfo->LineSpeed != currentLineSpeed) || (pCardInfo->LineSpeed == RtndMediaSpeedUnknown))
		{
			UpdateSystemTimeIncrement(pCardInfo);
		}
#endif
		// Update the current line speed
		currentLineSpeed = pCardInfo->LineSpeed;
	}

	return WaitStat;
}

// Must be called with locked transmit DMA
static void DrvCheckTxDescriptors(DRVDEV* pCardInfo, int QueNum, BOOL callAppCallback)
{
	PDRV_QUE_DESCRIPTORS pAdapterQ = &pCardInfo->adapterQ[QueNum];
	volatile ULONG TXDescriptorIndex = pAdapterQ->OldestUsedTxDescriptor;
	volatile IGC_ADV_TRANSMIT_DESCRIPTOR* pTransmitDescriptor = &pAdapterQ->TxDescriptors[TXDescriptorIndex];

	while (pTransmitDescriptor->wb.status & IGC_RXD_STAT_DD)
	{
		RTNAL_FRAME* pTxNalFrame = pAdapterQ->pTxNalFrames[TXDescriptorIndex];
		
		if (pTxNalFrame != NULL)
		{
			// If the NAL frame is used with this packet, call the callback

			// Timestamps in a packet are not supported at present
			pTxNalFrame->bTimeStampValid = FALSE;
			pTxNalFrame->timeStamp.nsec = 0;
			pTxNalFrame->timeStamp.sec = 0;

			pTxNalFrame->frameStatus = 0;  // no errors detected.  Not sure how to check

			if (!callAppCallback)
			{
				// Called in context of the driver, don't call application-level callback.
				pTxNalFrame->fpCallBack = NULL;
			}

			// Make the callback to NAL, which will:
			//		Call an application-level callback,
			//		Make a NAL frame available for transmit
			RtnTransmitCompleteCallback(pTxNalFrame);

			// Clear pointer to the NAL frame
			pAdapterQ->pTxNalFrames[TXDescriptorIndex] = NULL;
		}

		TXDescriptorIndex++;
		TXDescriptorIndex %= pAdapterQ->NumTxDescriptors;

		// Keep OldestUsedTxDescriptor, NumTxDescriptorsAvail synchronized with DD bit
		pAdapterQ->OldestUsedTxDescriptor = TXDescriptorIndex;
		pAdapterQ->NumTxDescriptorsAvail++;
		//RtPrintf("DrvCheckTxDescript: pAdapterQ->NumTxDescriptorsAvail %d \n", pAdapterQ->NumTxDescriptorsAvail);
		// Update thr descriptor
		pTransmitDescriptor->wb.status = 0;

		// go to next descriptor
		pTransmitDescriptor = &pAdapterQ->TxDescriptors[TXDescriptorIndex];
	}
}

// Driver Transmit Complete thread
// 
static ULONG WINAPI DrvTxCompleteThread(DRVDEV* pCardInfo)
{
	int QueNum;
	DWORD WaitStat = WAIT_FAILED;

	for (;;)
	{
		WaitStat = RtWaitForSingleObject(pCardInfo->hTxCompleteEvent, INFINITE);
		if (WaitStat != WAIT_OBJECT_0)
			break;

		for (QueNum = 0; QueNum < pCardInfo->numberOfTxNicQueues; QueNum++)
		{
			if ((pCardInfo->adapterQ[QueNum].flags & DRV_TX_NOTIFY_FLAG) == 0)
			{
				RtIgcEnterCriticalLock(&pCardInfo->csTxDMARing[QueNum]);

				// Don't call application-level tx complete callback in the driver context
				DrvCheckTxDescriptors(pCardInfo, QueNum, DONT_CALL_APP_LEVEL_CALLBACK);

				RtIgcLeaveCriticalLock(&pCardInfo->csTxDMARing[QueNum]);
			}
		}
	}
	return WaitStat;
}

/**--------------------------------------------------------------------------------------------------
\fn	RTNDPROC int RtndServiceTransmitQueue(void *ndp);

Called by NAL API library in application context to service transmit DMA ring
Must not be called this function in driver context!

\param ndp	- network queue pointer:
RtnGetDataLong(ndp, RTNAL_DRIVER_INDEX_OFFSET) must return driver interface index,
RtnGetDataLong(ndp, RTNAL_QUEUE_INDEX_OFFSET) must return interface queue index
\return 0 on success, -1 if ndp is invalid pointer
*-----------------------------------------------------------------------------------------------**/
RTNDPROC int
RtndServiceTransmitQueue(void* ndp)
{
	int errorCode = 0;
	DRVDEV* pCardInfo;
	ULONG drvIndex;
	ULONG queNum;

	if (ndp == NULL)
		return -1;

	drvIndex = RtnGetDataLong(ndp, RTNAL_DRIVER_INDEX_OFFSET);
	queNum = RtnGetDataLong(ndp, RTNAL_QUEUE_INDEX_OFFSET);

	if (drvIndex >= pGlobals->nCards)
		return -1;

	pCardInfo = pGlobals->drvCards[drvIndex];

	if (pCardInfo == NULL || (int)queNum >= pCardInfo->numberOfTxNicQueues)
	{
		DebugMsg(("%s pCardInfo == NULL in RtndServiceTransmitQueue\n", pGlobals->msgPrefix));
		return -1;
	}

	// TRUE for application-level callback from RtnTransmitCompleteCallback
	RtIgcEnterCriticalLock(&pCardInfo->csTxDMARing[queNum]);

	// Call application-level tx complete callback in application context
	DrvCheckTxDescriptors(pCardInfo, queNum, CALL_APP_LEVEL_CALLBACK);

	RtIgcLeaveCriticalLock(&pCardInfo->csTxDMARing[queNum]);
	return 0;
}


static DrvTerminateThreads(DRVDEV* pCardInfo)
{
	if ((pCardInfo->LinkStatus)
		&& (pCardInfo->hLineStatusThread != NULL))
	{
		TerminateThread(pCardInfo->hLineStatusThread, ~0);
		RtCloseHandle(pCardInfo->hLineStatusThread);
		pCardInfo->hLineStatusThread = NULL;
	}
	if (pCardInfo->hDrvTxCompleteThread != NULL)
	{
		TerminateThread(pCardInfo->hDrvTxCompleteThread, ~0);
		RtCloseHandle(pCardInfo->hDrvTxCompleteThread);
		pCardInfo->hDrvTxCompleteThread = NULL;
	}
}


static BOOL
DrvCreateThread(DRVDEV* pCardInfo, LPTHREAD_START_ROUTINE threaProc, DWORD idealProcessor, DWORD priority, char* threadName, HANDLE* phThread)
{
	BOOL Result = TRUE;
	DWORD	threadID;
	HANDLE	hThread;

	hThread = CreateThread(NULL,
		0,
		threaProc, // pointer to thread code
		pCardInfo, // pointer to card data
		CREATE_SUSPENDED,
		&threadID);
	*phThread = hThread;  // give the handle back to the caller

	if (hThread == NULL)
	{
		Result = FALSE;
		goto error_;
	}
	if (!RtSetThreadPriority(hThread, priority) || !RtSetThreadTimeQuantum(hThread, 0))
	{
		Result = FALSE;
		goto error_;
	}
	if (!SetThreadIdealProcessor(hThread, idealProcessor))
	{
		Result = FALSE;
		goto error_;
	}

	if (0xFFFFFFFF == ResumeThread(hThread))
	{
		Result = FALSE;
		goto error_;
	}

	//DebugMsg(("Thread Created:  Name=%s ThreadID=%d,  Priority=%d idealProcessor=%x\n", threadName, threadID, priority, idealProcessor));

error_:
	if (!Result)
	{
		if (hThread != NULL)
		{
			TerminateThread(hThread, ~0);
			RtCloseHandle(hThread);
			*phThread = NULL;  // give the handle back to the caller
		}
	}
	return Result;
}

HANDLE hDumpdesc;
static void DrvStopThreads(DRVDEV* pCardInfo)
{
	if (pCardInfo->hLineStatusThread != NULL)
	{
		TerminateThread(pCardInfo->hLineStatusThread, ~0U);
		RtCloseHandle(pCardInfo->hLineStatusThread);
		pCardInfo->hLineStatusThread = NULL;
		
		TerminateThread(hDumpdesc, ~0U);
		RtCloseHandle(hDumpdesc);
		hDumpdesc = NULL;
	}
	if (pCardInfo->hDrvTxCompleteThread != NULL)
	{
		TerminateThread(pCardInfo->hDrvTxCompleteThread, ~0U);
		RtCloseHandle(pCardInfo->hDrvTxCompleteThread);
		pCardInfo->hDrvTxCompleteThread = NULL;
	}
}


static BOOL DrvCreateThreads(DRVDEV* pCardInfo)
{
	BOOL Result = TRUE;
	

	if (pCardInfo->LinkStatus)
	{
		Result = DrvCreateThread(pCardInfo,
			DrvLineStatusThread,
			pCardInfo->LineStatusIdealProcessor,
			pCardInfo->LineStatusPriority,
			"DrvLineStatusThread",
			&pCardInfo->hLineStatusThread);

		if (!Result)
		{
			DebugMsg(("error creating driver DrvLineStatusThread thread - error: 0x%x\n", RtGetLastError()));
			goto error_;
		}
	}

	if (pCardInfo->RunTxCompleteThread)
	{
		Result = DrvCreateThread(pCardInfo,
			DrvTxCompleteThread,
			pCardInfo->TxCompleteIdealProcessor,
			pCardInfo->TxCompletePriority,
			"DrvTxCompleteThread",
			&pCardInfo->hDrvTxCompleteThread);

		if (!Result)
		{
			DebugMsg(("error creating driver DrvTxCompleteThread thread - error: 0x%x\n", RtGetLastError()));
			goto error_;
		}
	}

error_:

	if (!Result)
	{
		DrvStopThreads(pCardInfo);
	}
	return Result;
}




BOOL DrvProbePciVendorID(USHORT pci_vendor_id, USHORT pci_device_id)
{
	pci_vendor_info_t* ent;

	ent = igc_vendor_info_array;  // list of device ID's is a global array
	while (ent->pvi_vendor_id != 0)
	{
		if ((pci_vendor_id == ent->pvi_vendor_id) &&
			(pci_device_id == ent->pvi_device_id))
		{
			return(TRUE);
		}
		ent++;
	}

	return(FALSE);
}


/**--------------------------------------------------------------------------------------------------
   \fn	static BOOL DrvCheckForNetworkCard(PPCI_COMMON_CONFIG pPciInfo, ULONG Bus, PCI_SLOT_NUMBER SlotNumber, DRVDEV_INIT_PARAMS *pInitParams, BOOL fMacAddressPresent)

   \brief	DrvCheckForNetworkCard is called by the stack for each PCI/PCIe device it finds. If
			the PCI Vendor and Device ID's match to a card handled by this driver
			DrvCheckForNetworkCard returns true and the stack will stop calling
			DrvCheckForNetworkCard.

   \param	pPciInfo		  	Information describing the PCI.
   \param	Bus				  	The bus number.
   \param	SlotNumber		  	The slot number.
   \param	pInitParams		  	A  parameters list containing initialise parameters.
   \param	fMacAddressPresent	true if the MAC address present.

   \return	TRUE if a match is found.  FALSE indicates tells the stack to call again with the
			next PCI/PCIe device info.

*-----------------------------------------------------------------------------------------------**/
static BOOL DrvCheckForNetworkCard(PPCI_COMMON_CONFIG pPciInfo,
	ULONG Bus,
	PCI_SLOT_NUMBER SlotNumber,
	DRVDEV_INIT_PARAMS* pInitParams,
	BOOL fMacAddressPresent)
{
	DRVDEV* pCardInfo = pInitParams->pCardInfo;
	LARGE_INTEGER BusAddr, TranslatedAddr;
	ULONG AddressSpace;
	ULONG i, j;
	WORD wPort;
	DWORD dwTemp;
	PHYSICAL_ADDRESS PhysicalAddr;
	ULONG RegisterSize;

	pCardInfo->adapterType = 0;

	// device found, see if its the desired network card
	if ((pPciInfo->BaseClass == 0x02) && (pPciInfo->SubClass == 0x00) && (pPciInfo->ProgIf == 0x00))
	{
		if (DrvProbePciVendorID(pPciInfo->VendorID, pPciInfo->DeviceID))
		{
			pCardInfo->adapterType = pPciInfo->DeviceID;
		}

		if (pCardInfo->adapterType)
		{
			// The following will match the PCI bus and slot information with values specified in the interface 
			// configuration if present. Use of the init params avoids the problem where the MAC check fails later after
			// the card PCI information is updated. This breaks later checks for the PCI information since it
			// has been set with previously seen PCI information. The init params are never changed.
			if ((pInitParams->BusNumber <= PCI_MAX_BUS) &&
				(pInitParams->DeviceNumber <= PCI_MAX_DEVICES) &&
				(pInitParams->FuncNumber <= PCI_MAX_FUNCTION))
			{

				if (!((pInitParams->BusNumber == Bus) &&
					(pInitParams->DeviceNumber == SlotNumber.u.bits.DeviceNumber) &&
					(pInitParams->FuncNumber == SlotNumber.u.bits.FunctionNumber)))
				{
					SetLastError(ERROR_DEV_NOT_EXIST);
					return FALSE; // no match

				}
			}

			// Initialize basic information and get MAC type
			pCardInfo->Irq = pPciInfo->u.type0.InterruptLine;
			pCardInfo->BusNumber = Bus;
			pCardInfo->FuncNumber = SlotNumber.u.bits.FunctionNumber;
			pCardInfo->SlotNumber = SlotNumber;
			pCardInfo->CsrBase = NULL;
			pCardInfo->IOBase = 0;
			pCardInfo->VendorId = pPciInfo->VendorID;
			pCardInfo->DeviceId = pPciInfo->DeviceID;
			pCardInfo->RevId = pPciInfo->RevisionID;

			// First, find IO space base address if present (usually BAR2).
			for (j = PCI_BAR0; j <= PCI_BAR5; j++)

			{
				dwTemp = pPciInfo->u.type0.BaseAddresses[j];
				if (PCI_BAR_TYPE(dwTemp) == PCI_BAR_TYPE_MEM)
				{
					if (PCI_BAR_MEM_TYPE(dwTemp) == PCI_BAR_MEM_TYPE_64BIT)
						j++;
					continue;
				}

				BusAddr.QuadPart = pPciInfo->u.type0.BaseAddresses[j];
				AddressSpace = PCI_BAR_TYPE_IO;
				if (!RtTranslateBusAddress(PCIBus, Bus, BusAddr, &AddressSpace, &TranslatedAddr))
				{
					SetLastError(ERROR_DEVICE_ENUMERATION_ERROR);
					DebugMsg(("CheckForNetworkCard, error translating bus address (BAR=%d).\n", j));
					continue;
				}
				dwTemp = TranslatedAddr.LowPart; //pPciInfo->u.type0.BaseAddresses[1 - 5];
				if (PCI_BAR_TYPE(dwTemp) == PCI_BAR_TYPE_IO) // IO space address
				{
					// Found BAR with IO space address
					dwTemp &= 0x0000FFFC; // strip away unwanted bits
					wPort = ((WORD)dwTemp);
					pCardInfo->IOBase = (ULONG)wPort;
					if (pCardInfo->IOBase != 0)
					{
						break; // found
					}
				}
			}

			// Next locate the memory address for the NVM FLASH for ICH8 cards using BAR1. Other 
			// cards use EECD or EERD CSR registers to access EEPROM and do not need to use the 
			// FLASH device.

			// Finally, locate the memory registers (CSR) for the card using BAR0
			BusAddr.LowPart = pPciInfo->u.type0.BaseAddresses[0];
			if (PCI_BAR_MEM_TYPE(pPciInfo->u.type0.BaseAddresses[0]) == PCI_BAR_MEM_TYPE_64BIT)
			{
				BusAddr.HighPart = pPciInfo->u.type0.BaseAddresses[1];
			}
			else
			{
				BusAddr.HighPart = 0;
			}
			AddressSpace = PCI_BAR_TYPE_MEM;
			if (!RtTranslateBusAddress(PCIBus, Bus, BusAddr, &AddressSpace, &TranslatedAddr))
			{
				SetLastError(RT_ERROR_TRANSLATING_PCI_BUS_ADDRESS);
				DebugMsg(("CheckForNetworkCard, error translating bus CSR memory address (BAR0).\n"));
				return FALSE;
			}

			if (PCI_BAR_TYPE(TranslatedAddr.LowPart) == PCI_BAR_TYPE_MEM) // its a memory address
			{
				RegisterSize = INTEL_IGC_REGISTERMAP_SIZE;
				PhysicalAddr.QuadPart = (TranslatedAddr.QuadPart & 0xFFFFFFFFFFFFFFF0); // strip of non-address bits
				pCardInfo->CsrPhysical = PhysicalAddr;
				pCardInfo->CsrBase = (UCHAR*)RtMapMemory(PhysicalAddr, RegisterSize, MmNonCached);
			}
			if (pCardInfo->CsrBase == NULL)
			{
				SetLastError(RT_ERROR_MAPPING_REGISTERS);
				DebugMsg(("CheckForNetworkCard, error acquiring and mapping memory registers.\n"));
				return FALSE;
			}

			// Check for match with MAC address using the init params
			if (fMacAddressPresent)
			{
				// Case where address is specified, check init params against what was read from LAN NVM for the 
				// card. Note that even if the PCI information was specified and matched above, if the MAC was also
				// specified, it will be matched as well.
				for (i = 0; i < 6; i++)
				{
					if (pCardInfo->EnetAddr[i] != pInitParams->MacAddress[i])
					{
						RtUnmapMemory(pCardInfo->CsrBase);
						pCardInfo->CsrBase = NULL;
						SetLastError(RT_ERROR_MAC_ADDRESS_MISMATCH);
						DebugMsg(("DrvCheckForNetworkCard, fMacAddressPresent and LOC but MAC address does not match %x != %x \n", pCardInfo->EnetAddr[i], pInitParams->MacAddress[i]));
						return FALSE; // not matched, keep looking
					}
				}
			}
			// Else MAC address is not specified so any MAC address is fine. Note that the extra check for
			// lastbus etc was removed. This was in place to support finding a second port on a card with the
			// same MAC address. This was needed when only a MAC address could be used to specify which adapter.
			// The use of PCI information allows a particular port on a mulit-port card to be specified since
			// each port has a different function number.
			//DebugMsg(("CheckForNetworkCard. DEVICE=%x PCI=%2.2x:%2.2x:%2.2x MAC=%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\n", 
			//		pCardInfo->DeviceId,
			//		pCardInfo->BusNumber, SlotNumber.u.bits.DeviceNumber, SlotNumber.u.bits.FunctionNumber,
			//		pCardInfo->EnetAddr[0], pCardInfo->EnetAddr[1], pCardInfo->EnetAddr[2],
			//		pCardInfo->EnetAddr[3], pCardInfo->EnetAddr[4], pCardInfo->EnetAddr[5]));

#ifdef _DEBUG
			DebugMsg(("CheckForNetworkCard. DEVICE=%x PCI=%d;%d;%d\n",
				 pCardInfo->DeviceId,
				 pCardInfo->BusNumber, SlotNumber.u.bits.DeviceNumber, SlotNumber.u.bits.FunctionNumber));
#endif
			return TRUE; // found, stop looking
		}
	}

	SetLastError(ERROR_DEVICE_ENUMERATION_ERROR);

	return FALSE; // not found, keep going
}


static int DrvSplitLoc(char* string, char* retstr[], int size)
{
	int Num_of_Str;
	char** ret = retstr;
	char* s = string;
	char ch;

	if (*s == '\0')
		return(0);

	Num_of_Str = size;
	for (;;)
	{
		*ret++ = s;
		Num_of_Str--;
		if (Num_of_Str == 0)
			break;
		while ((ch = *s++) != ';')
			if (ch == '\0')
				return(size - Num_of_Str);
		*(s - 1) = '\0';
	}

	Num_of_Str = size;
	for (;;)
	{
		while ((ch = *s++) != ';')
			if (ch == '\0')
				return(Num_of_Str);
		Num_of_Str++;
	}

	return 0;
}


BOOL DrvGetInitParameters(PRTNAL_INTERFACE pNetInterface, DRVDEV_INIT_PARAMS* InitParams)
{
	BOOL fMACAddressPresent;
	char* LocStr[LOCS];
	char DefLocStr[LOCS * 2];
	char IntStr[LOCS];
	int		QueNum = 0;


	// Don't verify MAC address in interface configuration
	fMACAddressPresent = FALSE;

	// Load a default PCI location string
	_itoa((int)PCI_MAX_BUS + 1, IntStr, 10);
	strcpy(DefLocStr, IntStr);

	_itoa((int)PCI_MAX_DEVICES + 1, IntStr, 10);
	strcat(DefLocStr, ";");
	strcat(DefLocStr, IntStr);

	_itoa((int)PCI_MAX_FUNCTION + 1, IntStr, 10);
	strcat(DefLocStr, ";");
	strcat(DefLocStr, IntStr);

	if (pNetInterface->PciBusLocationString)
	{
		// Since defaults were loaded, this should always be true, load the init struct
		DrvSplitLoc(pNetInterface->PciBusLocationString, LocStr, LOCS);

		InitParams->BusNumber = atoi(LocStr[BUS_LOC]);
		InitParams->DeviceNumber = atoi(LocStr[DEV_LOC]);
		InitParams->FuncNumber = atoi(LocStr[FUNC_LOC]);

		if ((InitParams->BusNumber == 0) &&
			(InitParams->DeviceNumber == 0) &&
			(InitParams->FuncNumber == 0))
		{
			InitParams->BusNumber = PCI_MAX_BUS + 1;
			InitParams->DeviceNumber = PCI_MAX_DEVICES + 1;
			InitParams->FuncNumber = PCI_MAX_FUNCTION + 1;
		}
	}

	return FALSE;
}


/**--------------------------------------------------------------------------------------------------
   \fn	static void DrvSetupMtu(DRVDEV *pCardInfo, ULONG RequestedMtu)

   \brief	Set the MTU and frame size information for this adapter.

   \param [in,out]	pCardInfo	device information block.
   \param	RequestedMtu	 	MTU requested in interface configuration.

*-----------------------------------------------------------------------------------------------**/
static void DrvSetupMtu(DRVDEV* pCardInfo, ULONG RequestedMtu)
{
	ULONG FrameSize = RequestedMtu + (ETHERNET_HEADER_LEN + ETHERNET_FCS_LEN);

	// Check input values, default to standard ethernet frame size if input invalid
	if (FrameSize < ETHERNET_MIN_LEN)
	{
		DebugMsg(("DrvSetupMtu, invalid frame size: %d (using standard value %d)\n", FrameSize, ETHERNET_STD_LEN));
		FrameSize = ETHERNET_STD_LEN;
	}

	if (FrameSize > DRV_MAX_JUMBO_LEN)
	{

		DebugMsg(("DrvSetupMtu, invalid frame size: %d (using maximum jumbo value %d)\n", FrameSize, DRV_MAX_JUMBO_LEN));
		FrameSize = DRV_MAX_JUMBO_LEN;
	}

	// Set the values for the MTU and Rx buffer size. Note that the small frame sizes below 2048 are
	// not being used. Anything smaller than this will use the Rx buffer size 2048 (which is the size
	// for 1500 MTU frames).
	pCardInfo->Mtu = FrameSize - (ETHERNET_HEADER_LEN + ETHERNET_FCS_LEN);
	pCardInfo->MaxPacket = FrameSize;

	if (FrameSize <= DRV_RXBUFFER_2048)
	{
		pCardInfo->RxBufferSize = DRV_RXBUFFER_2048;
	}
	else
	{
		// Round up RxBufferSize to 1K
		pCardInfo->RxBufferSize = ROUNDUP(FrameSize, 1024);
	}
}

static ULONG GetNumberTxBuffers(ULONG requestedTxDescriptors)
{
	ULONG actualTxDescriptors = requestedTxDescriptors;

	if (actualTxDescriptors <= 0)
		actualTxDescriptors = DEFAULT_TXD;

	if (actualTxDescriptors < MIN_TXD)
		actualTxDescriptors = MIN_TXD;

	if (actualTxDescriptors > MAX_TXD)
		requestedTxDescriptors = MAX_TXD;

	// NumTxDescriptors must be a multiple of 8 (rounding up)
	actualTxDescriptors += REQ_TX_DESCRIPTOR_MULTIPLE - 1;
	actualTxDescriptors &= ~(REQ_TX_DESCRIPTOR_MULTIPLE - 1);

	return actualTxDescriptors;
}

static ULONG GetNumberRxBuffers(ULONG requestedRxDescriptors)
{
	ULONG actualRxDescriptors = requestedRxDescriptors;

	if (actualRxDescriptors <= 0)
		actualRxDescriptors = DEFAULT_RXD;

	if (actualRxDescriptors < MIN_RXD)
		actualRxDescriptors = MIN_RXD;

	if (actualRxDescriptors > MAX_RXD)
		actualRxDescriptors = MAX_RXD;

	// NumTxDescriptors must be a multiple of 8 (rounding up)
	actualRxDescriptors += REQ_RX_DESCRIPTOR_MULTIPLE - 1;
	actualRxDescriptors &= ~(REQ_RX_DESCRIPTOR_MULTIPLE - 1);

	return actualRxDescriptors;
}

static int MaxRxNicQuesForAdapterType(DWORD adapterType)
{
	return MAX_NUMBER_OF_RX_QUEUES;
}

static int MaxTxNicQuesForAdapterType(DWORD adapterType)
{
	return MAX_NUMBER_OF_TX_QUEUES;
}


// Sets interrupt moderation interval in nanoseconds
static int SetInterruptModeration(DRVDEV* pCardInfo, int interval)
{
	 /* Verify that interval (in ns) is between 0 and (2^13-1)*1000 */
	 /* Note: 0 means "No interrupt moderation" */
	 if (interval < 0 || interval > 8191000)
	 {
		  return ERROR_INVALID_PARAMETER;
	 }

	 /* Convert nanoseconds to microseconds (round up) */
	 const int us = (interval + 999) / 1000;

	 /* Prepare the value of the EITR register */
	 const __int32 regVal = 0           // b0..1: Reserved
		  | (us << 2)  // b2..14: Interval
		  | (0 << 15)  // b15: LLI_EN (disable Low Latency Interrupts)
		  | (0 << 16)  // b16..20: LL Counter
		  | (0 << 21)  // b21..30: Moderation Counter
		  | (0 << 31); // b31: CNT_INGR

/* Write the EITR[0] register */
/* Note: we assume here that multi vector MSI-X is not activated */
	 IGC_WRITE_REG(&pCardInfo->adapter.hw, IGC_EITR(0), regVal);

	 return ERROR_SUCCESS;
}

int
DrvGetSetupParameters(PRTNAL_INTERFACE pNetInterface, DRVDEV* pCardInfo)
{
	int		QueNum = 0;
	ULONG	ulNumXmitBuffers;
	ULONG	ulNumRecvBuffers;

#ifndef ETHERNET_HEADER_LEN
#define ETHERNET_HEADER_LEN 14
#endif

	if ((pNetInterface->IntPriority < 0) || (pNetInterface->IntPriority > 127))
	{
		if (pGlobals->displayErrorMessages)
		{
			DebugMsg(("interrupt priority must be between 0 and 127.\n"));
		}
		SetLastError(RT_INVALID_INTERRUPT_PRIORITY);

		return (-1);
	}

	// Force using MSI mode instead of MSI-X mode
	// until the problem with MSI-X mode is fixed.
	pNetInterface->InterruptType = ATTACH_MESSAGE_BASED_PREFER_MSI;

	pCardInfo->PreferMsi = pNetInterface->InterruptType == ATTACH_MESSAGE_BASED_PREFER_MSI;
	pCardInfo->LineBasedOnly = pNetInterface->InterruptType == ATTACH_LINE_BASED;

	pCardInfo->IntPriority = pNetInterface->IntPriority;
	pCardInfo->IntIdealProc = pNetInterface->IntIdealProcessor;
	pCardInfo->IntAffinityMask = 1i64 << pNetInterface->IntIdealProcessor;

	pCardInfo->LinkStatus = pNetInterface->LinkStatus;
	pCardInfo->LineStatusPriority = pNetInterface->LinkStatusPriority;
	pCardInfo->LineStatusIdealProcessor = pNetInterface->LinkStatusIdealProcessor;

	pCardInfo->RunTxCompleteThread = TRUE;
	pCardInfo->TxCompleteIdealProcessor = pNetInterface->TxCompleteIdealProcessor;
	pCardInfo->TxCompletePriority = pNetInterface->TxCompletePriority;

	ulNumXmitBuffers = GetNumberTxBuffers(pNetInterface->NumTxBuffers);
	ulNumRecvBuffers = GetNumberRxBuffers(pNetInterface->NumRxBuffers);

	pCardInfo->numberOfRxNicQueues = pNetInterface->NumberOfRxQueues;
	pCardInfo->numberOfTxNicQueues = pNetInterface->NumberOfTxQueues;

	if (pCardInfo->numberOfRxNicQueues == 0) {
		pCardInfo->numberOfRxNicQueues = 1;
	}
	else if (pCardInfo->numberOfRxNicQueues > MaxRxNicQuesForAdapterType(pCardInfo->adapterType)) {
		pCardInfo->numberOfRxNicQueues = MaxRxNicQuesForAdapterType(pCardInfo->adapterType);
	}

	if (pCardInfo->numberOfTxNicQueues == 0) {
		pCardInfo->numberOfTxNicQueues = 1;
	}
	else if (pCardInfo->numberOfTxNicQueues > MaxTxNicQuesForAdapterType(pCardInfo->adapterType)) {
		pCardInfo->numberOfTxNicQueues = MaxTxNicQuesForAdapterType(pCardInfo->adapterType);
	}

	// The driver only supports the default queue == 0
	pCardInfo->defaultQueue = 0;

	for (QueNum = 0; QueNum < pCardInfo->numberOfRxNicQueues; QueNum++)
	{
		pCardInfo->adapterQ[QueNum].NumRxDescriptors = ulNumRecvBuffers;
	}

	for (QueNum = 0; QueNum < pCardInfo->numberOfTxNicQueues; QueNum++)
	{
		pCardInfo->adapterQ[QueNum].NumTxDescriptors = ulNumXmitBuffers;
	}

	// Call routine to setup the MTU and frame/receive sizes
	DrvSetupMtu(pGlobals->drvCards[pGlobals->nCards], pNetInterface->MaxPacketSize - ETHERNET_HEADER_LEN);

#ifdef ZEROCOPY 
	pCardInfo->ZeroCopy = TRUE;
#else
	pCardInfo->ZeroCopy = FALSE;
#endif

	return(0);
}


/**--------------------------------------------------------------------------------------------------
\fn	static int DrvReceivePacket(DRVDEV *pCardInfo, void *ndp, unsigned int QueNum)

\brief	Called by RtndReceive to proccesss incoming packets.

\param [in,out]	pCardInfo	device information block.
\param [in]		ndp			Network pointer passed to RtndReceive
\param [in]		QueNum		Which queue to process.

\return	0 on success (if data copied), errorCode on failure

*-----------------------------------------------------------------------------------------------**/
static int DrvReceivePacket(DRVDEV* pCardInfo, void* ndp, ULONG QueNum)
{
	struct igc_hw* hw = &pCardInfo->adapter.hw;

	PDRV_QUE_DESCRIPTORS pAdapterQ = &pCardInfo->adapterQ[QueNum];
	volatile IGC_ADV_RECEIVE_DESCRIPTOR* CurrentRxDescriptor;
	PHYSICAL_ADDRESS physicalAddr;
	volatile int RxDescriptorIndex;

	int errorCode = 0;

	RtIgcEnterCriticalLock(&pCardInfo->csRxDMARing[QueNum]);

	if (!pCardInfo->DeviceIsUp)
	{
		RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[QueNum]);
		DebugMsg(("DrvReceivePacket, ignoring packets while device is not fully up - 1st instance. \n"));
		return ERROR_NO_NETWORK;
	}

	RxDescriptorIndex = pAdapterQ->NextRxDescriptorToCheck;
	CurrentRxDescriptor = &pAdapterQ->RxDescriptors[RxDescriptorIndex];

	// look at waiting packet
	if ((CurrentRxDescriptor->wb.upper.status_error & IGC_RXD_STAT_DD) == 0)
	{
		RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[QueNum]);
		return ERROR_NO_DATA;
	}

	if (!DRV_IS_RX_ATTACHED(pAdapterQ->flags))
	{
		errorCode = ERROR_DEVICE_NOT_CONNECTED;
	}
	else if ((CurrentRxDescriptor->wb.upper.status_error & IGC_RXD_STAT_EOP) == 0)
	{
		errorCode = ERROR_IO_DEVICE;
	}
	/* Make sure bad packets are discarded */
	else if (CurrentRxDescriptor->wb.upper.status_error & IGC_RXDEXT_ERR_FRAME_ERR_MASK)
	{
		// (RX data error) or (IPv4 checksum was calculated and error) or (TCP/UDP checksum was calculated and error)
		if ((CurrentRxDescriptor->wb.upper.status_error & IGC_RXDEXT_STATERR_RXE) ||
			((CurrentRxDescriptor->wb.upper.status_error & (IGC_RXDEXT_STATERR_IPE | IGC_RXD_STAT_IPCS)) ==
				(IGC_RXDEXT_STATERR_IPE | IGC_RXD_STAT_IPCS)) ||
			((CurrentRxDescriptor->wb.upper.status_error & (IGC_RXDEXT_STATERR_L4E | IGC_RXD_STAT_L4CS)) ==
				(IGC_RXDEXT_STATERR_L4E | IGC_RXD_STAT_L4CS))
			)
		{
#ifdef _DEBUG
			DebugMsg(("RtIGC: ReceivePacket: packet frame error.  status_error=%x\n", CurrentRxDescriptor->wb.upper.status_error));
#endif
			errorCode = ERROR_IO_DEVICE;
		}
	}

	if (errorCode == 0)
	{
		RTNAL_FRAME* pCurrentRxNalFrame = &pAdapterQ->RxNalFrames[RxDescriptorIndex];
		volatile unsigned char* rxPacket = pAdapterQ->rxFrameBuffVirtualAddr[RxDescriptorIndex];
		ULONG	Length = CurrentRxDescriptor->wb.upper.length;

		if (pCardInfo->ulLatencyRecvUpdated > 0 && Length > 1500)
		{
			LARGE_INTEGER liValue;

			liValue.QuadPart = (pCardInfo->ulLatencyRecvUpdated) * 10;  //convert to 100 nonsec.
			RtSleepFt(&liValue);
		}

		// TODO add code for timestamp in packet
		if (pCurrentRxNalFrame != NULL && pCurrentRxNalFrame->fpCallBack != NULL)
		{
			BOOL dropped;

			pCurrentRxNalFrame->frameSize = Length;
			pCurrentRxNalFrame->frameBufferVirtualAddr = (void*)rxPacket;

			dropped = pCurrentRxNalFrame->fpCallBack(pCurrentRxNalFrame);
			if (dropped)
				errorCode = ERROR_DISCARDED;
		}

		if (errorCode == 0)
		{
			// get a queued buffer from the stack to place data into
			void* mp = RtnGetPacket(ndp, Length);

			if (mp == NULL)
			{
				// Update bad stats
				pCardInfo->ComStats.Sbrecnt++;
#ifdef _DEBUG
				DebugMsg(("DrvReceivePacket, Unable to allocate receive buffer of size=%d.\n", Length));
#endif
				errorCode = ERROR_INVALID_USER_BUFFER;
			}
			else
			{
				unsigned long dataLength = 0;
				unsigned long* pRecvBuf = NULL; // a receive message buffer pointer
				unsigned long* tndp = NULL;

				RtnDecodePacket((void*)mp, &tndp, &pRecvBuf, &dataLength);

				if ((pRecvBuf == NULL) || (dataLength < Length))
				{
					// Update bad stats
					pCardInfo->ComStats.Sbrecnt++;
#ifdef _DEBUG
					DebugMsg(("DrvReceivePacket, unable decode receive buffer for Length %u, Returned: ptr %p, length %u.\n",
						Length, pRecvBuf, dataLength));
#endif
					errorCode = ERROR_INVALID_USER_BUFFER;
				}
				else
				{
					// Update good stats
					pCardInfo->ComStats.Sbrcnt++;

					// Copy the packet.
					memcpy(pRecvBuf, (void*)rxPacket, Length);
				}
			}
		}
	}

	physicalAddr = pCardInfo->adapterQ[QueNum].rxFrameBuffPhysicalAddr[RxDescriptorIndex];
	physicalAddr.LowPart &= ~1;
	CurrentRxDescriptor->read.pkt_addr = physicalAddr;
	CurrentRxDescriptor->read.hdr_addr = physicalAddr;

	IGC_WRITE_REG(&pCardInfo->adapter.hw, IGC_RDT(QueNum), RxDescriptorIndex);

	RxDescriptorIndex++;
	RxDescriptorIndex %= pCardInfo->adapterQ[QueNum].NumRxDescriptors;
	pCardInfo->adapterQ[QueNum].NextRxDescriptorToCheck = RxDescriptorIndex;
	if (pAdapterQ->RxDescriptorsNotified > 0)
		pAdapterQ->RxDescriptorsNotified--;

	RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[QueNum]);

	return errorCode;

}

static int DrvReceivePacketWithCallback(DRVDEV* pCardInfo, unsigned int QueNum)
{
	struct igc_hw* hw = &pCardInfo->adapter.hw;
	PDRV_QUE_DESCRIPTORS pAdapterQ = &pCardInfo->adapterQ[QueNum];
	volatile IGC_ADV_RECEIVE_DESCRIPTOR* CurrentRxDescriptor;
	PHYSICAL_ADDRESS physicalAddr;
	volatile int RxDescriptorIndex;
	int errorCode = 0;

	RtIgcEnterCriticalLock(&pCardInfo->csRxDMARing[QueNum]);

	if (!pCardInfo->DeviceIsUp)
	{
		RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[QueNum]);
		DebugMsg(("ReceivePackets, ignoring packets while device is not fully up - 1st instance. \n"));
		return ERROR_NO_NETWORK;
	}

	RxDescriptorIndex = pAdapterQ->NextRxDescriptorToCheck;
	CurrentRxDescriptor = &pAdapterQ->RxDescriptors[RxDescriptorIndex];

	// look at waiting packet
	if ((CurrentRxDescriptor->wb.upper.status_error & IGC_RXD_STAT_DD) == 0)
	{
		RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[QueNum]);
		return ERROR_NO_DATA;
	}

	if (pAdapterQ->pRxAppContext == NULL)
	{
		errorCode = ERROR_DEVICE_NOT_CONNECTED;
	}
	else if ((CurrentRxDescriptor->wb.upper.status_error & IGC_RXD_STAT_EOP) == 0)
	{
		errorCode = ERROR_IO_DEVICE;
	}
	else if (CurrentRxDescriptor->wb.upper.status_error & IGC_RXDEXT_ERR_FRAME_ERR_MASK)
	{
		// (RX data error) or (IPv4 checksum was calculated and error) or (TCP/UDP checksum was calculated and error)
		if ((CurrentRxDescriptor->wb.upper.status_error & IGC_RXDEXT_STATERR_RXE) ||
			((CurrentRxDescriptor->wb.upper.status_error & (IGC_RXDEXT_STATERR_IPE | IGC_RXD_STAT_IPCS)) ==
				(IGC_RXDEXT_STATERR_IPE | IGC_RXD_STAT_IPCS)) ||
			((CurrentRxDescriptor->wb.upper.status_error & (IGC_RXDEXT_STATERR_L4E | IGC_RXD_STAT_L4CS)) ==
				(IGC_RXDEXT_STATERR_L4E | IGC_RXD_STAT_L4CS))
			)
		{
#ifdef _DEBUG
			DebugMsg(("RtIGC: ReceivePacket: packet frame error.  status_error=%x\n", CurrentRxDescriptor->wb.upper.status_error));
#endif
			errorCode = ERROR_IO_DEVICE;
		}
	}
	if (errorCode == 0)
	{
		RTNAL_FRAME* pCurrentRxNalFrame = &pAdapterQ->RxNalFrames[RxDescriptorIndex];
		volatile unsigned char* rxPacket = pAdapterQ->rxFrameBuffVirtualAddr[RxDescriptorIndex];
		ULONG	Length = CurrentRxDescriptor->wb.upper.length;

		if (pCardInfo->ulLatencyRecvUpdated > 0 && Length > 1500)
		{
			LARGE_INTEGER liValue;

			liValue.QuadPart = (pCardInfo->ulLatencyRecvUpdated) * 10;  //convert to 100 nonsec.
			RtSleepFt(&liValue);
		}
		// TODO - add code for timestamp in packet
		if (pCurrentRxNalFrame != NULL && pCurrentRxNalFrame->fpCallBack != NULL)
		{
			pCurrentRxNalFrame->frameSize = Length;
			pCurrentRxNalFrame->frameBufferVirtualAddr = (void*)rxPacket;

			pCurrentRxNalFrame->fpCallBack(pCurrentRxNalFrame);
		}
		else
		{
			errorCode = ERROR_NOT_SUPPORTED;
		}
	}

	physicalAddr = pCardInfo->adapterQ[QueNum].rxFrameBuffPhysicalAddr[RxDescriptorIndex];
	physicalAddr.LowPart &= ~1;

	CurrentRxDescriptor->read.pkt_addr = physicalAddr;
	CurrentRxDescriptor->read.hdr_addr = physicalAddr;

	IGC_WRITE_REG(&pCardInfo->adapter.hw, IGC_RDT(QueNum), RxDescriptorIndex);

	RxDescriptorIndex++;
	RxDescriptorIndex %= pCardInfo->adapterQ[QueNum].NumRxDescriptors;
	pCardInfo->adapterQ[QueNum].NextRxDescriptorToCheck = RxDescriptorIndex;
	if (pAdapterQ->RxDescriptorsNotified > 0)
		pAdapterQ->RxDescriptorsNotified--;

	RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[QueNum]);

	return errorCode;
}

// Discard all receceived packets for the queue
// Assumes the NIC is configured
// Must call with locked receive DMA
static ULONG FlushReceiveQueue(DRVDEV* pCardInfo, ULONG queueNumber)
{
	PDRV_QUE_DESCRIPTORS pAdapterQ = &pCardInfo->adapterQ[queueNumber];
	DWORD RxDescriptorIndex = pAdapterQ->NextRxDescriptorToCheck;
	DWORD lastFlushed = RxDescriptorIndex;

	volatile IGC_ADV_RECEIVE_DESCRIPTOR* pRxDescriptor = &pAdapterQ->RxDescriptors[RxDescriptorIndex];
	ULONG queRecvCount = 0;
	struct igc_hw* hw = &pCardInfo->adapter.hw;
	
	// Flush DMA for all received packets, discarding them
	while ((pRxDescriptor->wb.upper.status_error & IGC_RXD_STAT_DD) &&
		(queRecvCount < pAdapterQ->NumRxDescriptors))
	{
		PHYSICAL_ADDRESS physicalAddr = pAdapterQ->rxFrameBuffPhysicalAddr[RxDescriptorIndex];
		physicalAddr.LowPart &= ~0x01;

		pRxDescriptor->read.pkt_addr = physicalAddr;
		pRxDescriptor->read.hdr_addr = physicalAddr;
	
		// The index to write to tail register after the loop
		lastFlushed = RxDescriptorIndex;

		// Increment packet count
		queRecvCount++;

		// Advance to the next buffer descriptor
		RxDescriptorIndex++;
		RxDescriptorIndex %= pAdapterQ->NumRxDescriptors;
		pRxDescriptor = &pAdapterQ->RxDescriptors[RxDescriptorIndex];
	} //while

	// If flushed some receive packets
	if (queRecvCount > 0)
	{
		// Advance the tail register 
		IGC_WRITE_REG(&pCardInfo->adapter.hw, IGC_RDT(queueNumber), lastFlushed);
		// Update the buffer descriptor index
		pAdapterQ->NextRxDescriptorToCheck = RxDescriptorIndex;

		// Set number of notified packets to 0
		pAdapterQ->RxDescriptorsNotified = 0;
	}

	// Return the number of discarded received packets
	return queRecvCount;
}

// Return the number of received packets in DMA for the queue
// Must call with locked receive DMA
static ULONG GetReceiveDMACount(DRVDEV* pCardInfo, unsigned int QueNum)
{
	PDRV_QUE_DESCRIPTORS pAdapterQ = &pCardInfo->adapterQ[QueNum];
	volatile ULONG RxDescriptorIndex = pAdapterQ->NextRxDescriptorToCheck;
	volatile IGC_ADV_RECEIVE_DESCRIPTOR* pCurrentRxDescriptor;

	ULONG fullCnt = 0;

	pCurrentRxDescriptor = &pAdapterQ->RxDescriptors[RxDescriptorIndex];

	while ((fullCnt < pAdapterQ->NumRxDescriptors) && (pCurrentRxDescriptor->wb.upper.status_error & IGC_RXD_STAT_DD) != 0)
	{
		fullCnt++;

		RxDescriptorIndex++;
		RxDescriptorIndex %= pAdapterQ->NumRxDescriptors;

		pCurrentRxDescriptor = &pAdapterQ->RxDescriptors[RxDescriptorIndex];
	}

	return fullCnt;
}

#pragma optimize( "", off )
static VOID RTAPI
DrvMessageBasedEnableInterrupt(DRVDEV* pCardInfo, ULONG MessageId, BOOLEAN bEnable)
{
	struct igc_hw* hw = &pCardInfo->adapter.hw;
	if (bEnable)
	{
		IGC_WRITE_REG(hw, IGC_EIMS, (1 << MessageId));
	}
	else
	{
		IGC_WRITE_REG(hw, IGC_EIMC, (1 << MessageId));
	}

	return;
}
#pragma optimize( "", on )


//-----------------------------------------------------------------------------
// Function:    DrvCommonMIST
//
// Purpose:     Card interrupt handler thread
//				
// Inputs:      pCardInfo - pointer to the card data structure
//              MessageId - MSIX message id, i.e., MSIX vector number
//
// Return:      TRUE, always handled
//
static BOOLEAN RTAPI
DrvCommonMIST(DRVDEV* pCardInfo, ULONG MessageId)
{
	struct igc_hw* hw = &pCardInfo->adapter.hw;
	ULONG RecvCount[MAX_NUMBER_OF_QUEUES] = { 0 };
	BOOL fLineStatusEvent = FALSE;
	DWORD IcrContents = 0;
	DWORD EicrContents = 0;
	BOOL messageBelongsToTxQueue = FALSE;
	BOOL messageBelongsToRxQueue = FALSE;
	ULONG  i = 0;
	int	  QueNum;
	u32   reg_exicr;

	//
	// Read Other Interrupt Causes (Clear by Read)
	//
	if (MessageId == 0)
	{
		IcrContents = IGC_READ_REG(hw, IGC_ICR);
		reg_exicr = IGC_READ_REG(hw, IGC_EICR);

//		RtPrintf("DrvCommonMIST IcrContents: 0x%x EICR 0%x Msg ID %d \n", IcrContents, reg_exicr, MessageId);
	}

	if (IcrContents != 0)
	{
		// If a LineStatus change has occured...
		if (IcrContents & (IGC_ICR_RXSEQ | IGC_ICR_LSC))
		{
			fLineStatusEvent = TRUE;
		}
	}

	// If message ID is mapped to a queue.
	if ((MessageId > 0) &&
		((MessageId < pCardInfo->usedMessagesCount) || (pCardInfo->usedMessagesCount == 1)))
	{

		// Figure out if message belongs to receive or transmit queue
		for (i = 0; i < pCardInfo->messageToInerruptKindQueueMap[MessageId].numberOfQueues; i++)
		{
			messageBelongsToTxQueue = (pCardInfo->messageToInerruptKindQueueMap[MessageId].queues[i].kind == InterruptKindTransmit) | messageBelongsToTxQueue;
			messageBelongsToRxQueue = (pCardInfo->messageToInerruptKindQueueMap[MessageId].queues[i].kind == InterruptKindReceive) | messageBelongsToRxQueue;

			// If we already know that it is Rx and Tx, no need to search more; 
			if (messageBelongsToTxQueue & messageBelongsToRxQueue)
			{
				break;
			}
		}
	}

	// For messages that belong to Rx queues. 
	if (TRUE == messageBelongsToRxQueue)
	{

		//RtPrintf("MID: %d is RX. QCNT: %d", MessageId, pCardInfo->messageToInerruptKindQueueMap[MessageId].numberOfQueues);

		// For each queue mapped for the current message
		for (i = 0; i < pCardInfo->messageToInerruptKindQueueMap[MessageId].numberOfQueues; i++)
		{
			PDRV_QUE_DESCRIPTORS pAdapterQ = NULL;

			if (pCardInfo->messageToInerruptKindQueueMap[MessageId].queues[i].kind != InterruptKindReceive)
			{
				continue;
			}

			QueNum = pCardInfo->messageToInerruptKindQueueMap[MessageId].queues[i].queue;

			//RtPrintf(" QNUM: %d\n", QueNum);

			// If data has been received...
			pAdapterQ = &pCardInfo->adapterQ[QueNum];

			if (DRV_IS_RX_POLLING(pAdapterQ->flags))
			{
				// Do nothing for polling applications.
				// Check is unlocked, but DMA will be flushed
				// when the application stops polling
				continue;
			}

			// Lock DMA for the RX queue
			RtIgcEnterCriticalLock(&pCardInfo->csRxDMARing[QueNum]);

			if (DRV_IS_RX_NOTIFY((pAdapterQ->flags)))
			{
				// An application is attached and wants to be notified of receive
				DWORD RxDescriptorIndex = pAdapterQ->NextRxDescriptorToCheck;
				ULONG queRecvCount = 0;

				// Count total received packets in DMA
				while ((pAdapterQ->RxDescriptors[RxDescriptorIndex].wb.upper.status_error & IGC_RXD_STAT_DD) &&
					(queRecvCount < pAdapterQ->NumRxDescriptors))
				{
					queRecvCount++;

					RxDescriptorIndex++;
					RxDescriptorIndex %= pAdapterQ->NumRxDescriptors;
				} // while

				if (queRecvCount >= pAdapterQ->RxDescriptorsNotified)
				{
					// Subtract already notified packets to
					// calculate new received packets to notify
					queRecvCount -= pAdapterQ->RxDescriptorsNotified;

					// Notified after IST return
					pAdapterQ->RxDescriptorsNotified += queRecvCount;
				}
				else
				{
					DebugMsg(("RtIGC: Invalid queRecvCount: queRecvCount %d, RxDescriptorsNotified %d\n", queRecvCount, pAdapterQ->RxDescriptorsNotified));
				}

				// Notify receive of packets, not notified before
				if (queRecvCount > 0)
				{
					if (pAdapterQ->bRxMonitor)
					{
						GenerateMonitorEvent(pCardInfo->EnetAddr, (BYTE)QueNum, RTNAL_MONITOR_EVENT_T1_RX, queRecvCount, 0);
					}
					RtnNotifyRecvQueue(pCardInfo->ndp, queRecvCount, QueNum);
				}

			} // DRV_IS_RX_NOTIFY
			else if (!DRV_IS_RX_POLLING(pAdapterQ->flags))
			{
				// An application is not attached or not polling at the moment
				// Flush DMA ring for this queue
				FlushReceiveQueue(pCardInfo, QueNum);

			} // !DRV_IS_RX_POLLING

			  // Unlock DMA for the RX queue
			RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[QueNum]);
		} // for 
	}

	if ((fLineStatusEvent) && (pCardInfo->hLineStatusEvent != NULL))
	{
		RtSetEvent(pCardInfo->hLineStatusEvent);
	}

	// For messages that belong to Tx queues. 
	if (TRUE == messageBelongsToTxQueue)
	{
		BOOL notifyDriverTxCompleteThread = FALSE;

		for (i = 0; i < pCardInfo->messageToInerruptKindQueueMap[MessageId].numberOfQueues; i++)
		{
			PDRV_QUE_DESCRIPTORS pAdapterQ = NULL;

			// If this mapping is not for Tx - keep going.
			if (pCardInfo->messageToInerruptKindQueueMap[MessageId].queues[i].kind != InterruptKindTransmit)
			{
				continue;
			}

			// Queue number for the current mapping.
			QueNum = pCardInfo->messageToInerruptKindQueueMap[MessageId].queues[i].queue;

			pAdapterQ = &pCardInfo->adapterQ[QueNum];

			// Notify the application of transmitted packets,
			// if the application has configured transmit notifications for this queue
			if ((pAdapterQ->flags & DRV_TX_NOTIFY_FLAG) == 0 ||
				!RtnNotifyTransmitQueue(pCardInfo->ndp, QueNum))
			{
				// an application was not notified for this queue
				notifyDriverTxCompleteThread = TRUE;
			}
		}

		if (notifyDriverTxCompleteThread)
		{
			// Notify the driver transmit complete thread,
			// if an application was not notified for some queue
			RtSetEvent(pCardInfo->hTxCompleteEvent);
		}

	}

	if ((fLineStatusEvent) && (pCardInfo->hLineStatusEvent != NULL))
	{
		// Notify driver line status thread of link change
		RtSetEvent(pCardInfo->hLineStatusEvent);
	}

	return TRUE;
}

//-----------------------------------------------------------------------------
// Function:    DrvMessageBasedMISR
//
// Purpose:     Card interrupt handler service routine for MSIX based interrupt
//              handling.
//				
// Inputs:      pCardInfo - pointer to the card data structure
//              MessageId - MSIX message id, i.e., MSIX vector number
//
// Return:      CallInterruptThread to call IST
//
static INTERRUPT_DISPOSITION RTAPI
DrvMessageBasedMISR(DRVDEV* pCardInfo, ULONG MessageId)
{
	return CallInterruptThread;
}
/**--------------------------------------------------------------------------------------------------
   \fn	static INTERRUPT_DISPOSITION RTAPI DrvCommonISR(DRVDEV *pCardInfo)

   \brief	When the interrupt occurs, the ISR calls the DrvCommonISR since it was supplied in
			the  RtAttachInterrupt call. This optional routine should be used to determine if the
			interrupt should be handled by this instance of the driver. Since it is called at
			interrupt level, this routine should complete its work quickly; 10-20 usec.

   \param [in,out]	pCardInfo	device information block.

   \return	\bPassToNextDevice if its associated device did not generate the interrupt.

			\bCallInterruptThread if its associated device generated the interrupt.  RTX will
			then schedule DrvCommonIST.

			\bDismiss if its associated device generated the interrupt and no further action is
			required of the handling thread (presumably because this routine has done all that is
			necessary).

*-----------------------------------------------------------------------------------------------**/
static INTERRUPT_DISPOSITION RTAPI  DrvCommonISR(DRVDEV* pCardInfo)
{

	// The driver could be built with just an IST but the ISR is included just in case
	// it is needed int the future


	return CallInterruptThread;
}

/**--------------------------------------------------------------------------------------------------
   \fn	static BOOLEAN RTAPI DrvCommonIST(DRVDEV *pCardInfo)

   \brief	DrvCommonIST is run by the Interrupt Service Thread after DrvCommonISR returns
			CallInterruptThread.

   \param [in,out]	pCardInfo	device information block.

   \return	If the interrupt came from its device returns TRUE.

			If its device did not generate the interrupt, the handling routine returns FALSE.

*-----------------------------------------------------------------------------------------------**/

static BOOLEAN RTAPI DrvCommonIST(DRVDEV* pCardInfo)
{
	struct igc_hw* hw = &pCardInfo->adapter.hw;
	u32   reg_eicr;
	u32   reg_exicr;
	BOOL  fReceiveEvent = FALSE;
	BOOL  fTransmitEvent = FALSE;
	BOOL  fLineStatusEvent = FALSE;

	reg_eicr = IGC_READ_REG(hw, IGC_ICR);

	reg_exicr = IGC_READ_REG(hw, IGC_EICR);

	if (!reg_eicr)
	{
		Rt_igc_if_intr_enable(pCardInfo);
		return(FALSE);
	}
	do
	{
		//RtPrintf("In IST reg_eicr: 0x%x ext 0%x \n", reg_eicr, reg_exicr);

		if ((reg_eicr & IGC_ICR_RXT0) 
			|| (reg_eicr & IGC_ICR_RXDMT0))
			//|| (reg_exicr & IGC_EICR_RX_QUEUE0)
			//|| (reg_exicr & IGC_EICR_RX_QUEUE1)
			//|| (reg_exicr & IGC_EICR_RX_QUEUE2)
			//|| (reg_exicr & IGC_EICR_RX_QUEUE3))
		{
			// rx interrupt
			fReceiveEvent = TRUE;
		}

		if ((reg_eicr & IGC_ICR_TXDW)
			|| (reg_exicr & IGC_EICR_TX_QUEUE0)
			|| (reg_exicr & IGC_EICR_TX_QUEUE1)
			|| (reg_exicr & IGC_EICR_TX_QUEUE2)
			|| (reg_exicr & IGC_EICR_TX_QUEUE3))
		{
			// tx interrupt
			fTransmitEvent = TRUE;
		}

		if (reg_eicr & (IGC_ICR_RXSEQ | IGC_ICR_LSC))
		{
			// Link status change
			fLineStatusEvent = TRUE;
		}
	} while (reg_eicr = IGC_READ_REG(hw, IGC_ICR));

	Rt_igc_if_intr_enable(pCardInfo);

	if (fReceiveEvent)
	{

		int	  QueNum;

		// check how many rx packets are in
		for (QueNum = 0; QueNum < pCardInfo->numberOfRxNicQueues; QueNum++)
		{
			DRV_QUE_DESCRIPTORS* pAdapterQ = &pCardInfo->adapterQ[QueNum];

			if (DRV_IS_RX_POLLING(pAdapterQ->flags))
			{
				// Do nothing for polling applications.
				// Check is unlocked, but DMA will be flushed
				// when the application stops polling
				continue;
			}

			// Lock DMA for the receive queue
			RtIgcEnterCriticalLock(&pCardInfo->csRxDMARing[QueNum]);

			if (DRV_IS_RX_NOTIFY((pAdapterQ->flags)))
			{
				// An application is attached and wants to be notified of receive
				DWORD RxDescriptorIndex = pAdapterQ->NextRxDescriptorToCheck;
				ULONG queRecvCount = 0;
	

				// Count total receive packets in DMA
				while ((pAdapterQ->RxDescriptors[RxDescriptorIndex].wb.upper.status_error & IGC_RXD_STAT_DD)
					&& (queRecvCount < pAdapterQ->NumRxDescriptors))
				{
					queRecvCount++;

					RxDescriptorIndex++;
					RxDescriptorIndex %= pAdapterQ->NumRxDescriptors;
				} // while

				if (queRecvCount >= pAdapterQ->RxDescriptorsNotified)
				{
					// Subtract already notified packets to
					// calculate new received packets to notify
					queRecvCount -= pAdapterQ->RxDescriptorsNotified;

					// Notified after IST return
					pAdapterQ->RxDescriptorsNotified += queRecvCount;
				}
				else
				{
					DebugMsg(("Invalid queRecvCount: queRecvCount %d, RxDescriptorsNotified %d\n", queRecvCount, pAdapterQ->RxDescriptorsNotified));
				}

				// Notify receive of packets, not notified before
				if (queRecvCount > 0)
				{
					if (pAdapterQ->bRxMonitor)
					{
						GenerateMonitorEvent(pCardInfo->EnetAddr, (BYTE)QueNum, RTNAL_MONITOR_EVENT_T1_RX, queRecvCount, 0);
					}
					RtnNotifyRecvQueue(pCardInfo->ndp, queRecvCount, QueNum);
				}

			} // DRV_IS_RX_NOTIFY
			else if (!DRV_IS_RX_POLLING(pAdapterQ->flags))
			{
				// An application is not attached or not pollint at the moment
				// Flush DMA ring for this queue
				FlushReceiveQueue(pCardInfo, QueNum);

			} // !DRV_IS_RX_POLLING

			// Unlock DMA for the RX queue
			RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[QueNum]);
		} // for 
	} // if (fRecvEvent)

	// signal to check if any tx packets are done
	if (fTransmitEvent)
	{
		BOOL notifyDriverTxCompleteThread = FALSE;
		int QueNum;

		for (QueNum = 0; QueNum < pCardInfo->numberOfTxNicQueues; QueNum++)
		{
			DRV_QUE_DESCRIPTORS* pAdapterQ = &pCardInfo->adapterQ[QueNum];

			// Notify the application of transmitted packets,
			// if the application has configured transmit notifications for this queue
			if ((pAdapterQ->flags & DRV_TX_NOTIFY_FLAG) == 0 ||
				!RtnNotifyTransmitQueue(pCardInfo->ndp, QueNum))
			{
				// an application was not notified for this queue
				notifyDriverTxCompleteThread = TRUE;
			}
		}

		if (notifyDriverTxCompleteThread)
		{
			// Notify the driver transmit complete thread,
			// if an application was not notified for some queue
			RtSetEvent(pCardInfo->hTxCompleteEvent);
		}
	}

	// Link status change
	if (fLineStatusEvent && pCardInfo->hLineStatusEvent != NULL)
		RtSetEvent(pCardInfo->hLineStatusEvent);

	return(TRUE);
}

BOOL DrvAllocateTransmitFrameBuffers(DRVDEV* pCardInfo, int QueNum)
{
	volatile IGC_ADV_TRANSMIT_DESCRIPTOR* CurrentTxDescriptor;
	DRV_QUE_DESCRIPTORS* pAdapterQ = &pCardInfo->adapterQ[QueNum];
	ULONG TxBufferSize = pCardInfo->RxBufferSize;
	BYTE* pFrameBuffer;
	PHYSICAL_ADDRESS PhysAddr;
	unsigned int i;

	// Allocate transmit DMA buffers
	pFrameBuffer = DrvAllocateMemory(TRUE, TxBufferSize * pCardInfo->adapterQ[QueNum].NumTxDescriptors);
	if (pFrameBuffer == NULL)
	{
		DebugMsg(("RtndConfigure error.  DrvAllocateMemory for transmit DMA buffers failed\n"));
		return (FALSE);
	}

	PhysAddr = RtGetPhysicalAddress(pFrameBuffer);

	for (i = 0; i < pAdapterQ->NumTxDescriptors; i++) {

		pAdapterQ->txFrameBuffVirtualAddr[i] = pFrameBuffer;
		pAdapterQ->txFrameBuffPhysicalAddr[i] = PhysAddr;

		CurrentTxDescriptor = &pCardInfo->adapterQ[QueNum].TxDescriptors[i];
		CurrentTxDescriptor->read.buffer_addr = PhysAddr;

		pFrameBuffer += TxBufferSize;
		PhysAddr.QuadPart += TxBufferSize;
	}

	return TRUE;
}

BOOL DrvAllocateReceiveFrameBuffers(DRVDEV* pCardInfo, int QueNum)
{

	IGC_ADV_RECEIVE_DESCRIPTOR* CurrentDescriptor;
	DRV_QUE_DESCRIPTORS* pAdapterQ = &pCardInfo->adapterQ[QueNum];
	BYTE* pFrameBuffer;
	PHYSICAL_ADDRESS physicalAddr;
	ULONG RxBufferSize = pCardInfo->RxBufferSize + 16; // Add 16 byte padding for in-packet time stamps
	unsigned int i;

	pFrameBuffer = DrvAllocateMemory(TRUE, RxBufferSize * pCardInfo->adapterQ[QueNum].NumRxDescriptors);
	if (!pFrameBuffer)
	{
		return(FALSE);
	}

	physicalAddr = RtGetPhysicalAddress(pFrameBuffer);

	for (i = 0; i < pCardInfo->adapterQ[QueNum].NumRxDescriptors; i++)
	{
		CurrentDescriptor = &pAdapterQ->RxDescriptors[i];

		pCardInfo->adapterQ[QueNum].rxFrameBuffVirtualAddr[i] = pFrameBuffer;
		pCardInfo->adapterQ[QueNum].rxFrameBuffPhysicalAddr[i] = physicalAddr;

		CurrentDescriptor->read.pkt_addr = physicalAddr;
		// no header splitting so put the buffer address as the header
		CurrentDescriptor->read.hdr_addr = physicalAddr;

		pFrameBuffer += RxBufferSize;
		physicalAddr.QuadPart += RxBufferSize;
	}

	return TRUE;
}

BOOL DrvAllocateTransmitStructures(DRVDEV* pCardInfo, int QueNum)
{
	DWORD dwBlock;
	int size;

	// NumTxDescriptors is already initialized in RtndInitializeInterface

	dwBlock = 16;
	size = ROUNDUP(pCardInfo->adapterQ[QueNum].NumTxDescriptors * sizeof(IGC_ADV_TRANSMIT_DESCRIPTOR), 4096);
	dwBlock += size;

	// Allocate memory for transmit buffer descriptors
	pCardInfo->adapterQ[QueNum].pTxDescriptorBufferUnAligned = DrvAllocateMemory(TRUE, dwBlock);
	if (pCardInfo->adapterQ[QueNum].pTxDescriptorBufferUnAligned == NULL)
	{
		DebugMsg(("RtndConfigure error.  DrvAllocateMemory for pTxDescriptorBufferUnAligned failed\n"));
		return (FALSE);
	}

	pCardInfo->adapterQ[QueNum].pbyTxDescriptors = DrvPageAlignMemory(pCardInfo->adapterQ[QueNum].pTxDescriptorBufferUnAligned, &pCardInfo->adapterQ[QueNum].TxDescDMA);

	// Allocate memory for array of physical addresses to driver-allocated transmit data buffers
	pCardInfo->adapterQ[QueNum].txFrameBuffPhysicalAddr = (PHYSICAL_ADDRESS*)DrvAllocateMemory(FALSE, pCardInfo->adapterQ[QueNum].NumTxDescriptors * sizeof(DWORD*));
	if (pCardInfo->adapterQ[QueNum].txFrameBuffPhysicalAddr == NULL)
	{
		DebugMsg(("RtndConfigure error.  DrvAllocateMemory for txFrameBuffPhysicalAddr failed\n"));
		return (FALSE);
	}

	// Allocate memory for array of virtual addresses to driver-allocated transmit data buffers
	pCardInfo->adapterQ[QueNum].txFrameBuffVirtualAddr = (BYTE**)DrvAllocateMemory(FALSE, pCardInfo->adapterQ[QueNum].NumTxDescriptors * sizeof(DWORD*));
	if (pCardInfo->adapterQ[QueNum].txFrameBuffVirtualAddr == NULL)
	{
		DebugMsg(("RtndConfigure error.  DrvAllocateMemory for txFrameBuffVirtualAddr failed\n"));
		return (FALSE);
	}

	// Allocate memory for array of pointers to transmit NAL frames
	pCardInfo->adapterQ[QueNum].pTxNalFrames = (PRTNAL_FRAME*)DrvAllocateMemory(FALSE, pCardInfo->adapterQ[QueNum].NumTxDescriptors * sizeof(PRTNAL_FRAME));
	if (pCardInfo->adapterQ[QueNum].pTxNalFrames == NULL)
	{
		DebugMsg(("RtndConfigure error.  DrvAllocateMemory for pTxNalFrames failed\n"));
		return (FALSE);
	}

	pCardInfo->adapterQ[QueNum].TxDescriptors = (IGC_ADV_TRANSMIT_DESCRIPTOR*)pCardInfo->adapterQ[QueNum].pbyTxDescriptors;
	memset(pCardInfo->adapterQ[QueNum].TxDescriptors, 0, size);

	pCardInfo->adapterQ[QueNum].NumTxDescriptorsAvail = pCardInfo->adapterQ[QueNum].NumTxDescriptors;
	pCardInfo->adapterQ[QueNum].NextAvailTxDescriptor = 0;
	pCardInfo->adapterQ[QueNum].OldestUsedTxDescriptor = 0;

	return (TRUE);
}

BOOL DrvAllocateReceiveStructures(DRVDEV* pCardInfo, int QueNum)
{
	DWORD dwBlock;
	int size;

	// NumRxDescriptors is already initialized in RtndInitializeInterface

	dwBlock = 16;
	size = ROUNDUP(pCardInfo->adapterQ[QueNum].NumRxDescriptors * sizeof(IGC_ADV_RECEIVE_DESCRIPTOR), 4096);
	dwBlock += size;

	// Allocate memory for receive buffer descriptors
	pCardInfo->adapterQ[QueNum].pRxDescriptorBufferUnAligned = DrvAllocateMemory(TRUE, dwBlock);
	if (pCardInfo->adapterQ[QueNum].pRxDescriptorBufferUnAligned == NULL)
	{
		DebugMsg(("RtndConfigure error.  DrvAllocateMemory for pRxDescriptorBufferUnAligned failed\n"));
		return (FALSE);
	}
	pCardInfo->adapterQ[QueNum].pbyRxDescriptors = DrvPageAlignMemory(pCardInfo->adapterQ[QueNum].pRxDescriptorBufferUnAligned, &pCardInfo->adapterQ[QueNum].RxDescDMA);

	// Allocate memory for array of physical addresses to driver-allocated receive data buffers
	pCardInfo->adapterQ[QueNum].rxFrameBuffPhysicalAddr = (PHYSICAL_ADDRESS *)DrvAllocateMemory(FALSE, pCardInfo->adapterQ[QueNum].NumRxDescriptors * sizeof(DWORD*));
	if (pCardInfo->adapterQ[QueNum].rxFrameBuffPhysicalAddr == NULL)
	{
		DebugMsg(("RtndConfigure error.  DrvAllocateMemory for rxFrameBuffPhysicalAddr failed\n"));
		return (FALSE);
	}

	// Allocate memory for array of virtual addresses to driver-allocated receive data buffers
	pCardInfo->adapterQ[QueNum].rxFrameBuffVirtualAddr = (BYTE**)DrvAllocateMemory(FALSE, pCardInfo->adapterQ[QueNum].NumRxDescriptors * sizeof(DWORD*));
	if (pCardInfo->adapterQ[QueNum].rxFrameBuffVirtualAddr == NULL)
	{
		DebugMsg(("RtndConfigure error.  DrvAllocateMemory for rxFrameBuffVirtualAddr failed\n"));
		return (FALSE);
	}

	// Allocate memory for array of receive NAL frames
	pCardInfo->adapterQ[QueNum].RxNalFrames = (RTNAL_FRAME*)DrvAllocateMemory(FALSE, pCardInfo->adapterQ[QueNum].NumRxDescriptors * sizeof(RTNAL_FRAME));
	if (pCardInfo->adapterQ[QueNum].RxNalFrames == NULL)
	{
		DebugMsg(("RtndConfigure error.  DrvAllocateMemory for RxNalFrames failed\n"));
		return (FALSE);
	}

	pCardInfo->adapterQ[QueNum].RxDescriptors = (IGC_ADV_RECEIVE_DESCRIPTOR*)pCardInfo->adapterQ[QueNum].pbyRxDescriptors;
	
	memset(pCardInfo->adapterQ[QueNum].RxDescriptors, 0, size);

	pCardInfo->adapterQ[QueNum].NextRxDescriptorToCheck = 0;
	pCardInfo->adapterQ[QueNum].NextRxDescriptorToFill = 0;
	pCardInfo->adapterQ[QueNum].RxDescriptorsNotified = 0;

	return (TRUE);
}

BOOL DrvSetupReceiveStructures(DRVDEV* pCardInfo)
{
	int QueNum = 0;

	for (QueNum = 0; QueNum < pCardInfo->numberOfRxNicQueues; QueNum++)
	{
		// Allocate arrays for storing addresses
		if (!DrvAllocateReceiveStructures(pCardInfo, QueNum))
		{
			return FALSE;
		}

		// Allocate the actual frame buffers
		if (!DrvAllocateReceiveFrameBuffers(pCardInfo, QueNum))
		{
			return FALSE;
		}
	}

	return(TRUE);
}

BOOL DrvSetupTransmitStructures(DRVDEV* pCardInfo)
{
	int QueNum = 0;

	for (QueNum = 0; QueNum < pCardInfo->numberOfTxNicQueues; QueNum++)
	{
		// Allocate arrays for storing addresses
		if (!DrvAllocateTransmitStructures(pCardInfo, QueNum))
		{
			return FALSE;
		}

		// Allocate the actual frame buffers
		if (!DrvAllocateTransmitFrameBuffers(pCardInfo, QueNum))
		{
			return FALSE;
		}

	}
	return(TRUE);
}

static BOOL SetupInterruptMV(DRVDEV* pCardInfo)
{
	ATTACH_INTERRUPT_PARAMETERS AttachParams;
	ULONG MsgTableSize = 0;
	ULONG currentMessage = 0;
	ULONG messagesCount = pCardInfo->usedMessagesCount;

	// There must be at leat two messages
	if (messagesCount < 2)
	{
		messagesCount = 2;
	}

	AttachParams.AttachVersion = ATTACH_MESSAGE_BASED_MULTI_VECTOR;
	MsgTableSize = sizeof(INTERRUPT_MESSAGE_INFO) + (messagesCount - 1) * sizeof(INTERRUPT_MESSAGE_INFO_ENTRY);

	AttachParams.MessageBasedMvector.InterruptMessageTable =
		(PINTERRUPT_MESSAGE_INFO)DrvAllocateMemory(FALSE, MsgTableSize);

	AttachParams.MessageBasedMvector.InterruptMessageTable->MessageCount = messagesCount;
	AttachParams.MessageBasedMvector.pThreadAttributes = NULL;           // Security attributes (default)
	AttachParams.MessageBasedMvector.StackSize = 0;                      // Use default value
	AttachParams.MessageBasedMvector.pRoutine = DrvCommonMIST;           // MIST
	AttachParams.MessageBasedMvector.MyInterrupt = DrvMessageBasedMISR;  // MISR
	AttachParams.MessageBasedMvector.EnableInterruptFromSpecifiedMessage = DrvMessageBasedEnableInterrupt; // EnableIntrrupt callback
	AttachParams.MessageBasedMvector.Context = pCardInfo;                // Context passed to IST, ISR, and EnableIntrrupt callback
	AttachParams.MessageBasedMvector.BusNumber = pCardInfo->BusNumber;   // Bus number
	AttachParams.MessageBasedMvector.SlotNumber = pCardInfo->SlotNumber; // Slot number																					
																	 //1 << RtGetCurrentProcessorNumber();                       // Rtss Processor Mask

	if (!ConfigureMsixMultiVectorLayout(pCardInfo))
	{
		DebugMsg(("%s error configuring MSIX multiple vectors layout.\n", pGlobals->msgPrefix));
		return FALSE;
	}

	// Define each message
	// Queue count can be larger than amount of messages.
	for (currentMessage = 0; currentMessage < messagesCount; currentMessage++)
	{

		AttachParams.MessageBasedMvector.InterruptMessageTable->MessageInfo[currentMessage].MessageId = currentMessage;
		AttachParams.MessageBasedMvector.InterruptMessageTable->MessageInfo[currentMessage].Priority = pCardInfo->IntPriority;  // IST priority

		if (currentMessage == 0)
			AttachParams.MessageBasedMvector.InterruptMessageTable->MessageInfo[currentMessage].TargetProcessor = pCardInfo->IntIdealProc;
		else
			AttachParams.MessageBasedMvector.InterruptMessageTable->MessageInfo[currentMessage].TargetProcessor = pCardInfo->messageToInerruptKindQueueMap[currentMessage].processor;

		// Keep for debugging
		//RtPrintf(
		//	"I225 SetupInterruptMV. Message: %d of %d. CORE: %d \n", currentMessage, messagesCount,
		//	AttachParams.MessageBasedMvector.InterruptMessageTable->MessageInfo[currentMessage].TargetProcessor
		//	);

	}

	pCardInfo->UsingMSI = TRUE;

	pCardInfo->hInterrupt = RtAttachInterrupt(&AttachParams);
	if (pCardInfo->hInterrupt == NULL)
	{
		SetLastError(RT_ERROR_FAILED_TO_ATTACH_INTERRUPT);
		DebugMsg(("I225: SetupInterruptMV error attaching IRQ., error %u\n", GetLastError()));
		return FALSE;
	}

	if (AttachParams.AttachVersion == ATTACH_MESSAGE_BASED_MULTI_VECTOR)
		DrvFreeMemory(FALSE, (BYTE*)AttachParams.MessageBasedMvector.InterruptMessageTable);

	return TRUE;
}

BOOL DrvAttachInterrupt(DRVDEV* pCardInfo)
{
	ULONG Capability;
	ULONG test = 0;
	DWORD tctl = 0;
	ATTACH_INTERRUPT_PARAMETERS AttachParams; // Attach interrupt parameters


	Capability = RtQueryPciMsiCapability(pCardInfo->BusNumber, pCardInfo->SlotNumber);

	if ((ULONG)(pCardInfo->LineBasedOnly) > Capability)
	{
		SetLastError(RT_ERROR_INCORRECT_INTERRUPT_TYPE);
		DebugMsg(("DrvAttachInterrupt, Incorrect Interrupt Type.\n"));
		return FALSE;
	}

	// Select line-based attach if device does not support MSI or input for line-based
	if ((pCardInfo->LineBasedOnly) || (!Capability))
	{
		AttachParams.AttachVersion = ATTACH_LINE_BASED;             // Line based version
		AttachParams.LineBased.pThreadAttributes = NULL;            // Security attributes (default)
		AttachParams.LineBased.StackSize = 0;                       // Use default value
		AttachParams.LineBased.pRoutine = DrvCommonIST;             // IST
		AttachParams.LineBased.Context = pCardInfo;                 // Context passed to IST and ISR
		AttachParams.LineBased.Priority = pCardInfo->IntPriority;   // IST priority
		AttachParams.LineBased.InterfaceType = PCIBus;              // PCI bus
		AttachParams.LineBased.BusNumber = pCardInfo->BusNumber;    // Bus number
		AttachParams.LineBased.SlotNumber = pCardInfo->SlotNumber;  // Slot number
		AttachParams.LineBased.BusInterruptLevel = pCardInfo->Irq;  // Interrupt level
		AttachParams.LineBased.BusInterruptVector = pCardInfo->Irq; // Interrupt vector
		AttachParams.LineBased.Shared = FALSE;                      // Not shared
		AttachParams.LineBased.InterruptMode = LevelSensitive;      // Type of interrupt
		AttachParams.LineBased.MyInterrupt = NULL;                  // ISR
		AttachParams.LineBased.ProcessorEnableMask = pCardInfo->IntAffinityMask; //Interrupt Ideal Processor affinity
		//1 << RtGetCurrentProcessorNumber();                     // Rtss Processor Mask
		pCardInfo->UsingMSI = FALSE;
		pCardInfo->adapter.intr_type = IFLIB_INTR_LEGACY;
	}
	else
	{
		if (pCardInfo->PreferMsi)
		{
			AttachParams.AttachVersion = ATTACH_MESSAGE_BASED_PREFER_MSI;            // Message based version with MSI preference
			pCardInfo->adapter.intr_type = IFLIB_INTR_MSI;
		}
		else
		{
			AttachParams.AttachVersion = ATTACH_MESSAGE_BASED;            // Message based version
			pCardInfo->adapter.intr_type = IFLIB_INTR_MSIX;
		}


			AttachParams.MessageBased.pThreadAttributes = NULL;           // Security attributes (default)
			AttachParams.MessageBased.StackSize = 0;                      // Use default value
			AttachParams.MessageBased.pRoutine = DrvCommonIST;            // IST
			AttachParams.MessageBased.Context = pCardInfo;                // Context passed to IST and ISR
			AttachParams.MessageBased.Priority = pCardInfo->IntPriority;  // IST priority
			AttachParams.MessageBased.BusNumber = pCardInfo->BusNumber;   // Bus number
			AttachParams.MessageBased.SlotNumber = pCardInfo->SlotNumber; // Slot number
			AttachParams.MessageBased.MyInterrupt = NULL;				  // DrvCommonISR;   // ISR
			AttachParams.MessageBased.ProcessorEnableMask = pCardInfo->IntAffinityMask; //Interrupt Ideal Processor affinity
			//1 << RtGetCurrentProcessorNumber();                       // Rtss Processor Mask
			pCardInfo->UsingMSI = TRUE;
	}

	//DebugMsg(("ATTACH-INTERRUPT TYPE=%s PCI=%2.2x:%2.2x:%2.2x\n", 
	//		(AttachParams.AttachVersion == ATTACH_LINE_BASED) ? "LINE" : "MESSAGE",
	//		pCardInfo->BusNumber, pCardInfo->SlotNumber.u.bits.DeviceNumber, 
	//		pCardInfo->SlotNumber.u.bits.FunctionNumber));
	// Attach interrupt

#ifdef MULTIVECTOR_SUPPORT
	if (AttachParams.AttachVersion != ATTACH_MESSAGE_BASED)
	{
		pCardInfo->hInterrupt = RtAttachInterrupt(&AttachParams);
	}
	else
	{
		//pCardInfo->hInterrupt = RtAttachInterrupt(&AttachParams);
		SetupInterruptMV(pCardInfo);
	}
#else
	pCardInfo->hInterrupt = RtAttachInterrupt(&AttachParams);
#endif

	if (pCardInfo->hInterrupt == NULL)
	{
		SetLastError(RT_ERROR_FAILED_TO_ATTACH_INTERRUPT);
		DebugMsg(("DrvAttachInterrupt, error attaching IRQ.\n"));
		return FALSE;
	}

	return(TRUE);
}

/**--------------------------------------------------------------------------------------------------
   \fn	static BOOL NicAdapterStart(DRVDEV *pCardInfo)

   \brief	Called to coordinate actually bringing up the hardware.

   \param [in,out]	pCardInfo	device information block.

   \return	TRUE if it succeeds, FALSE if it fails.

*-----------------------------------------------------------------------------------------------**/

static BOOL NicAdapterStart(DRVDEV* pCardInfo)
{
	ULONG test = 0;
	DWORD tctl = 0;
	int Vector = 0;
	int		QueNum = 0;
	int error = 0;
	int j;
	
	// Initialize critical section used to synchronize adapter accesses.
	for (QueNum = 0; QueNum < pCardInfo->numberOfRxNicQueues; QueNum++)
	{
		RtIgcInitializeCriticalLock(&pCardInfo->csRxDMARing[QueNum], "RxDMACriticalLock");
	}

	for (QueNum = 0; QueNum < pCardInfo->numberOfTxNicQueues; QueNum++)
	{
		RtIgcInitializeCriticalLock(&pCardInfo->csTxDMARing[QueNum], "TxDMACriticalLock");
	}
	RtIgcInitializeCriticalLock(&pCardInfo->csIoctl, "IoctlCriticalLock");

	// Create the transmit complete event
	pCardInfo->hTxCompleteEvent = RtCreateEvent(NULL, FALSE, FALSE, NULL);
	if (!pCardInfo->hTxCompleteEvent)
	{
		DebugMsg(("NicAdapterStart, error creating hTxCompleteEvent event.\n"));
		return (FALSE);
	}

	pCardInfo->hLineStatusEvent = NULL;

	if (pCardInfo->LinkStatus)
	{
		// Create the line status change event
		pCardInfo->hLineStatusEvent = RtCreateEvent(NULL, FALSE, FALSE, NULL);
		if (!pCardInfo->hLineStatusEvent)
		{
			DebugMsg(("NicAdapterStart, error creating line status event.\n"));
			return (FALSE);
		}
	}

	// Create the line event threads
	if (!DrvCreateThreads(pCardInfo))
	{
		DebugMsg(("NicAdapterStart, error creating receive and line event threads.\n"));
		return(FALSE);
	}

	// Call our equivalent of ixgbe_attach()
	if (igc_if_attach_pre(pCardInfo))
	{
		DebugMsg(("NicAdapterStart, igc_if_attach_pre errors\n"));
		return FALSE;
	}

	if (pCardInfo->LineBasedOnly)
	{
		pCardInfo->adapter.intr_type = IFLIB_INTR_LEGACY;
	}
	else if (pCardInfo->PreferMsi)
	{
		pCardInfo->adapter.intr_type = IFLIB_INTR_MSI;
	}
	else
	{
		pCardInfo->adapter.intr_type = IFLIB_INTR_MSIX;
	}


	// Setup and allocate memory for Tx/Rx DMA ring buffers
	if (!DrvSetupTransmitStructures(pCardInfo)
		|| !DrvSetupReceiveStructures(pCardInfo))
	{
		DebugMsg(("NicAdapterStart, DrvSetupTransmitStructures | DrvSetupReceiveStructures errors\n"));
		return FALSE;
	}

	igc_if_attach_post(pCardInfo);

	Rt_igc_if_init(pCardInfo);

	// Attach Interrupt
	if (!DrvAttachInterrupt(pCardInfo))
	{
		DebugMsg(("NicAdapterStart, DrvAttachInterrupt failed\n"));
		return(FALSE);
	}

	// Copy MAC address out of igc_hw hardware structure
	// pCardInfo->adapter.hw.mac.addr[j] was filled in by the igc_read_mac_addr call
	for (j = 0; j < ETH_ADDR_LEN; j++)
		pCardInfo->EnetAddr[j] = pCardInfo->adapter.hw.mac.addr[j];

	Rt_igc_if_intr_enable(pCardInfo);

#ifdef TIMESTAMP_SUPPORT
	NicPtpInit(pCardInfo);
#endif
	return(TRUE);
}


/**--------------------------------------------------------------------------------------------------
   \fn	static BOOL NicAdapterStop(DRVDEV *pCardInfo)

   \brief	Called by RtndUpDown() to coordinate actually bringing down the hardware.

   \param [in,out]	pCardInfo	device information block.

   \return	TRUE if it succeeds, FALSE if it fails.

*-----------------------------------------------------------------------------------------------**/
static BOOL NicAdapterStop(DRVDEV* pCardInfo)
{
	BOOL nStatus;
	struct igc_hw* hw = &pCardInfo->adapter.hw;

	int QueNum;

	// disable Interrupts
	Rt_igc_if_intr_disable(pCardInfo);
   
	// General Cleanup routines
	igc_if_stop(pCardInfo);
	igc_if_shutdown(pCardInfo);
	igc_if_detach(pCardInfo);

	Sleep(500);

	if (pCardInfo->hInterrupt != NULL)
	{
		if (!RtReleaseInterrupt(pCardInfo->hInterrupt))
		{
			DebugMsg(("NicAdapterStop, error detaching Interrupt.\n"));
			nStatus = FALSE;
		}
		else
		{
			pCardInfo->hInterrupt = NULL;
		}
	}

	Sleep(500);

	for (QueNum = 0; QueNum < pCardInfo->numberOfRxNicQueues; QueNum++)
	{
		RtIgcDeleteCriticalLock(&pCardInfo->csRxDMARing[QueNum]);
	}

	for (QueNum = 0; QueNum < pCardInfo->numberOfTxNicQueues; QueNum++)
	{
		RtIgcDeleteCriticalLock(&pCardInfo->csTxDMARing[QueNum]);
	}

	RtIgcDeleteCriticalLock(&pCardInfo->csIoctl);

	// release threads and events
	DrvTerminateThreads(pCardInfo);

	Sleep(500);

	if (pCardInfo->hTxCompleteEvent)
	{
		RtCloseHandle(pCardInfo->hTxCompleteEvent);
		pCardInfo->hTxCompleteEvent = NULL;
	}

	if (pCardInfo->hLineStatusEvent)
	{
		RtCloseHandle(pCardInfo->hLineStatusEvent);
		pCardInfo->hLineStatusEvent = NULL;
	}

	//  Free up memory
	for (QueNum = 0; QueNum < pCardInfo->numberOfRxNicQueues; QueNum++)
	{
		// DMA buffers were allocated as a single memory block
		DrvFreeMemory(TRUE, pCardInfo->adapterQ[QueNum].rxFrameBuffVirtualAddr[0]);
		DrvFreeMemory(TRUE, pCardInfo->adapterQ[QueNum].pRxDescriptorBufferUnAligned);
		DrvFreeMemory(FALSE, pCardInfo->adapterQ[QueNum].rxFrameBuffPhysicalAddr);
		DrvFreeMemory(FALSE, pCardInfo->adapterQ[QueNum].rxFrameBuffVirtualAddr);
		DrvFreeMemory(FALSE, pCardInfo->adapterQ[QueNum].RxNalFrames);
	}

	for (QueNum = 0; QueNum < pCardInfo->numberOfTxNicQueues; QueNum++)
	{
		// DMA buffers were allocated as a single memory block
		DrvFreeMemory(TRUE, pCardInfo->adapterQ[QueNum].txFrameBuffVirtualAddr[0]);
		DrvFreeMemory(TRUE, pCardInfo->adapterQ[QueNum].pTxDescriptorBufferUnAligned);
		DrvFreeMemory(FALSE, pCardInfo->adapterQ[QueNum].txFrameBuffPhysicalAddr);
		DrvFreeMemory(FALSE, pCardInfo->adapterQ[QueNum].txFrameBuffVirtualAddr);
		DrvFreeMemory(FALSE, pCardInfo->adapterQ[QueNum].pTxNalFrames);
	}

	RtUnmapMemory(pCardInfo->CsrBase);

	return(TRUE);
}

/**--------------------------------------------------------------------------------------------------
   \fn	RTNDPROC int RtndUpDown(void *ndp, unsigned short flags, char *options)

   \brief	RtndUpDown is called when the stack requires that the driver start (or stop) a
			particular NIC card. Given a flag value of non-zero RtndUpDown will bring up the NIC.
			An interrupt should be attached to a vector and enabled from the card; DMA should be
			turned on, etc. Given a flag value of zero RtndUpDown will bring down the NIC card.
			All interrupts should be disabled from the card and interrupt vector should be
			released. DMA should be turned off, and the card should be placed in a safe dormant
			state. Event handles should be closed, the Receive thread and possibly the Transmit
			thread should be terminated, etc.

   \param [in,out]	ndp	   	Network Device Pointer. Opaque network device identifier.
   \param	flags		   	Flag indicating if the network device is being started or stopped. 1
							means bring device up, 0 means bring device down.
   \param [in,out]	options	Reserved.

   \return	A return value of 0 indicates Success, -1 indicates Failure.

*-----------------------------------------------------------------------------------------------**/
RTNDPROC int RtndUpDown(void* ndp, unsigned short flags, char* options)
{
	DRVDEV* pCardInfo;
	int nStatus = 0;
	ULONG drvIndex;

	if (ndp == NULL)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	drvIndex = RtnGetDataLong(ndp, RTNAL_DRIVER_INDEX_OFFSET);
	if (drvIndex >= pGlobals->nCards)
	{
		SetLastError(ERROR_DEVICE_ENUMERATION_ERROR);
		return -1;
	}

	pCardInfo = pGlobals->drvCards[drvIndex];
	if (pCardInfo == NULL)
	{
		SetLastError(ERROR_DEV_NOT_EXIST);
		return -1;
	}

	if (flags) // GOING UP
	{
		//DebugMsg(("Device with Interface name %s has MAC address %0X:%0X:%0X:%0X:%0X:%0X\n", RtnGetDeviceName(ndp), pCardInfo->EnetAddr[0], pCardInfo->EnetAddr[1], pCardInfo->EnetAddr[2], pCardInfo->EnetAddr[3], pCardInfo->EnetAddr[4], pCardInfo->EnetAddr[5]));
		RtnSetLinkAddress(ndp, pCardInfo->EnetAddr, 6, pCardInfo->adapterQ[0].NumTxDescriptors - 1);
	}
	else // GOING DOWN
	{

		pCardInfo->DeviceIsUp = FALSE;

		if (!NicAdapterStop(pCardInfo))
		{
			DebugMsg(("RtndUpDown error.  NicAdapterStop failed\n"));
			return(-1);
		}
	}

	return 0;
}

/**--------------------------------------------------------------------------------------------------
\fn	RTNDPROC int RtndReceive(void *ndp)

\brief	RtndReceive is called by a NAL application receive thread.  The NAL application is
notified of a waiting receive packet when the IST routine calls RtnNotifyRecvQueue().
After that the application calls RtndReceive() presumably from a lower priority thread than
the IST.

RtndReceive calls a receive packed callback if it was provided in RtndAttachToReceiveQueue().
If a good paket was receive in DMA and not rejected by the callback, RtndReceive() calls
RtnGetPacket to retrieve a packet from ndp and RtnDecodePacket to retrieve a data pointer
to copy data to.

Return value 0 indicates that the data was copied to the application/stack buffer.

Possible error codes:
ERROR_BAD_ARGUMENTS			- ndp is NULL or invalid driver index or queue index was retrieved from ndp by RtnGetDataLong
ERROR_NO_NETWORK			- driver not initialized
ERROR_DEVICE_NOT_CONNECTED	- device queue is not attached to an application
ERROR_NO_DATA				- no received packet waitiin in DMA
ERROR_IO_DEVICE				- packet receive error
ERROR_INVALID_USER_BUFFER	- RtnGetPacket or RtnDecodePacket did not provide a valid user packet
ERROR_DISCARDED				- packet processed in a callback, the callback returned TRUE

\param [in,out]	ndp	Network Device Pointer. Opaque network device identifier.

\return	A return value of 0 indicates Success, -1 indicates Failure, use GetLastError() to get an error code.
*-----------------------------------------------------------------------------------------------**/
RTNDPROC int RtndReceive(void* ndp)
{
	DRVDEV* pCardInfo;
	int errorCode = 0;
	ULONG drvIndex;
	ULONG queNum;

	if (ndp == NULL) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	drvIndex = RtnGetDataLong(ndp, RTNAL_DRIVER_INDEX_OFFSET);
	queNum = RtnGetDataLong(ndp, RTNAL_QUEUE_INDEX_OFFSET);

	if (drvIndex >= pGlobals->nCards) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	pCardInfo = pGlobals->drvCards[drvIndex];

	if (pCardInfo == NULL || queNum >= (ULONG)pCardInfo->numberOfRxNicQueues)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	errorCode = DrvReceivePacket(pCardInfo, ndp, queNum); // receive a packet

	if (errorCode != 0)
	{
		SetLastError(errorCode);
		errorCode = -1;
	}

	return errorCode;

}

/**--------------------------------------------------------------------------------------------------
\fn	RTNDPROC int RtndReceiveWithCallback(ULONG interfaceNum, ULONG queueNumber)

\brief	RtndReceiveWithCallback is called by a NAL application receive thread.  The NAL application is
notified of a waiting receive packet when the IST routine calls RtnNotifyRecvQueue().
After that the application calls RtndReceiveWithCallback() presumably from a lower priority thread than
the IST.

RtndReceiveWithCallback calls a receive packed callback if it was provided in RtndAttachToReceiveQueue().

Return value 0 indicates that the receive packed callback was called, regardless of a callback return.

Possible error codes:
ERROR_BAD_ARGUMENTS			- ndp is NULL or invalid driver index or queue index was retrieved from ndp by RtnGetDataLong
ERROR_NO_NETWORK			- driver not initialized
ERROR_DEVICE_NOT_CONNECTED	- device queue is not attached to an application
ERROR_NO_DATA				- no received packet waitiin in DMA
ERROR_IO_DEVICE				- packet receive error
ERROR_NOT_SUPPORTED			- receive callback not configured

\param [in,out]	ndp	Network Device Pointer. Opaque network device identifier.

\return	A return value of 0 indicates Success, -1 indicates Failure , use GetLastError() to get an error code
*-----------------------------------------------------------------------------------------------**/
RTNDPROC int RtndReceiveWithCallback(ULONG interfaceNum, ULONG queueNumber)
{
	DRVDEV* pCardInfo;
	int errorCode = 0;

	if (interfaceNum >= pGlobals->nCards) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	// get pointer to the driver data structure
	pCardInfo = pGlobals->drvCards[interfaceNum];

	if (pCardInfo == NULL || queueNumber >= (ULONG)pCardInfo->numberOfRxNicQueues) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	errorCode = DrvReceivePacketWithCallback(pCardInfo, queueNumber);
	if (errorCode != 0)
	{
		SetLastError(errorCode);
	}

	return errorCode;
}


/**--------------------------------------------------------------------------------------------------
   \fn	RTNDPROC int RtndTransmit(void *mp)

   \brief	RtndTransmit is called when the calling application requires that the driver transmits a packet
			(pointed to by the parameter mp). The packet is in a structure format required by the
			calling application, which the driver does not need to know. The driver should use the
			RtnDecodePacket routine to obtain the network device (ndp), data pointer and data
			length for the packet.

			RtndTransmit copied application data to a driver's buffer

   \param [in,out]	mp	Pointer to a Packet to transmit.

   \return	A return value of 0 indicates Success, -1 indicates Failure.

*-----------------------------------------------------------------------------------------------**/
//static unsigned long cnt = 0;
RTNDPROC int RtndTransmit(void* mp)
{

	DRVDEV* pCardInfo;
	unsigned long* ndp;
	unsigned long* dataPtr;
	unsigned long dataLength;
	volatile IGC_ADV_TRANSMIT_DESCRIPTOR* CurrentTxDescriptor;
	volatile int TXDescriptorIndex;
	ULONG	drvIndex;
	ULONG	QueNum;
	BOOL	bTimeStamp = 0;

	if (mp == NULL)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	// 
	// get length of packet and pointer to data
	//
	RtnDecodePacket((void*)mp, &ndp, &dataPtr, &dataLength);

	// Check values, returned by RtnDecodePacket
	if (ndp == NULL)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	if (dataPtr == NULL)
	{
		SetLastError(ERROR_INVALID_ADDRESS);
		return -1;
	}

	drvIndex = RtnGetDataLong(ndp, RTNAL_DRIVER_INDEX_OFFSET);
	if (drvIndex >= pGlobals->nCards)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	//
	// get pointer to the driver data structure
	//
	pCardInfo = pGlobals->drvCards[drvIndex];

	if (pCardInfo == NULL)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	QueNum = RtnGetDataLong(ndp, RTNAL_QUEUE_INDEX_OFFSET);
	if (QueNum >= (ULONG)pCardInfo->numberOfTxNicQueues)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	// If the packet to be sent is invalid...
	if ((dataLength <= 0) || (dataLength > (pCardInfo->MaxPacket - ETHERNET_FCS_LEN)))
	{
		SetLastError(ERROR_INCORRECT_SIZE);
		return -1;
	}
	//RtPrintf("RtndTransmit(%d) datalen 0x%x \n", cnt++, dataLength);

	if (pCardInfo->hLineStatusEvent != NULL &&
		pCardInfo->LinkIsActive == RtndMediaConnectStateDisConnected)
	{
		SetLastError(ERROR_IO_DEVICE);
		return -1;
	}

	// Lock access to Tx buffer descriptors
	RtIgcEnterCriticalLock(&pCardInfo->csTxDMARing[QueNum]);

	//  Since wrap around will lock the NIC we can never use the last descriptor and 
	//  therefore we always need at least 2 descriptors.

	if (pCardInfo->adapterQ[QueNum].NumTxDescriptorsAvail < 2)
	{
		RtIgcLeaveCriticalLock(&pCardInfo->csTxDMARing[QueNum]);
#ifdef _DEBUG
		DebugMsg(("RtndTransmit failed.  No DMA descriptors remaining. Try increasing Transmit Buffers"));
#endif
		//increment no buffer count
		if (pCardInfo->ComStats.SbXnbc != 0xffffffff)
			pCardInfo->ComStats.SbXnbc++;

		SetLastError(ERROR_DEVICE_NO_RESOURCES);
		return -1;
	}

	// Grab the Index for the available TX descriptor.
	TXDescriptorIndex = pCardInfo->adapterQ[QueNum].NextAvailTxDescriptor;

	// RtndTransmit does not use NAL frames.
	// Clear the NAL frame pointer to let CheckTxDescriptors know that this is a driver-allocated packet
	pCardInfo->adapterQ[QueNum].pTxNalFrames[TXDescriptorIndex] = NULL;

	// Grab the actual TX descriptor.
	CurrentTxDescriptor = &pCardInfo->adapterQ[QueNum].TxDescriptors[TXDescriptorIndex];

	// Copy the buffer into the DMA buffer
	memcpy(pCardInfo->adapterQ[QueNum].txFrameBuffVirtualAddr[TXDescriptorIndex], dataPtr, dataLength);

	CurrentTxDescriptor->read.buffer_addr = pCardInfo->adapterQ[QueNum].txFrameBuffPhysicalAddr[TXDescriptorIndex];

	CurrentTxDescriptor->read.cmd_type_len = 0xffff & dataLength;

	/* Set basic descriptor constants */
	CurrentTxDescriptor->read.cmd_type_len |= IGC_ADVTXD_DTYP_DATA;
	CurrentTxDescriptor->read.cmd_type_len |= IGC_ADVTXD_DCMD_IFCS;
	CurrentTxDescriptor->read.cmd_type_len |= IGC_ADVTXD_DCMD_DEXT;
	CurrentTxDescriptor->read.cmd_type_len |= IGC_ADVTXD_DCMD_RS;
	CurrentTxDescriptor->read.cmd_type_len |= IGC_ADVTXD_DCMD_EOP;

	// CurrentTxDescriptor->config.typeAndLength |= RTIGB_ADVTXD_DCMD_VLE; to enable VLAN insertion 
#ifdef TIMESTAMP_SUPPORT
	if (pCardInfo->adapterQ[QueNum].bTimeStamp == TRUE)
		CurrentTxDescriptor->config.typeAndLength |= IGC_ADVTXD_MAC_TSTAMP;
#endif

	//	CurrentTxDescriptor->config.paylen  = RTIGB_TXD_DMA_TXDWB;  // report the DMA time of the tx desc wb
	//ToDo 
	//CurrentTxDescriptor->config.paylen &= IXGBE_ADVTXD_PAYLEN_MASK;
	CurrentTxDescriptor->read.olinfo_status |= dataLength << IGC_ADVTXD_PAYLEN_SHIFT;

	// Advance software descriptor index
	pCardInfo->adapterQ[QueNum].NextAvailTxDescriptor++;
	pCardInfo->adapterQ[QueNum].NextAvailTxDescriptor %= pCardInfo->adapterQ[QueNum].NumTxDescriptors;

	// Advance HW Tx Tail Pointer
	IGC_WRITE_REG(&pCardInfo->adapter.hw, IGC_TDT(QueNum), pCardInfo->adapterQ[QueNum].NextAvailTxDescriptor);

	// Update number of available descriptors
	pCardInfo->adapterQ[QueNum].NumTxDescriptorsAvail--;
	
	if (pCardInfo->adapterQ[QueNum].bTxMonitor)
	{
		GenerateMonitorEvent(pCardInfo->EnetAddr, (BYTE)QueNum, RTNAL_MONITOR_EVENT_T2_TX, 1, 0);
	}

	RtIgcLeaveCriticalLock(&pCardInfo->csTxDMARing[QueNum]);

	return 0;
}

/**--------------------------------------------------------------------------------------------------
\fn	RTNDPROC ULONG RtndTransmitEx(ULONG interfaceNumber, ULONG queueNumber, PRTNAL_FRAME pNalFrameArray[], ULONG numberOfFrames)

\brief	RtndTransmitEx is called when the application requires that the driver transmit an array of NAL frames. Pointers to NAL frames
	passed in pNalFrameArray. Each NAL frame has has virtual and physical addresses of the Ethernet frame to teransmit.
	RtndTransmitEx is a zero copy API.
	The driver does not have to check the frame length, because the caller - NAL - checks the length.

\param	interfaceNumber				[in] Driver interface index to transmit
\param	queueNumber					[in] Driver transmit queue index for this interface
\param	pNalFrameArray					[in] Array of pointers to frames to transmit
\param	numberOfFrames				[in] Number of packets to teransmit

\return	Number of transmitted packets. This function must set last error if the return value is less than numberOfFrames

The function can set one of the following error codes:
ERROR_BAD_ARGUMENTS - bad argument is passed in
ERROR_INVALID_ADDRESS - A pointer to a NAL frame is NULL
ERROR_IO_DEVICE		- link is down
ERROR_DEVICE_NO_RESOURCES - There are not ehough available buffer descriptors to send all of the requested packets
*-----------------------------------------------------------------------------------------------**/
RTNDPROC ULONG RtndTransmitEx(ULONG interfaceNumber, ULONG queueNumber, PRTNAL_FRAME pNalFrameArray[], ULONG numberOfFrames)
{
	DRVDEV* pCardInfo;
	PDRV_QUE_DESCRIPTORS pAdapterQ;
	ULONG bufferIndex = 0;
	DWORD errorCode = 0;

	if (numberOfFrames == 0 || interfaceNumber >= pGlobals->nCards) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return 0;
	}

	// get pointer to the driver data structure
	pCardInfo = pGlobals->drvCards[interfaceNumber];

	if (pCardInfo == NULL || queueNumber >= (ULONG)pCardInfo->numberOfTxNicQueues) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return 0;
	}

	if (pCardInfo->hLineStatusEvent != NULL &&
		pCardInfo->LinkIsActive == RtndMediaConnectStateDisConnected)
	{
		SetLastError(ERROR_IO_DEVICE);
		return 0;
	}

	pAdapterQ = &pCardInfo->adapterQ[queueNumber];

	RtIgcEnterCriticalLock(&pCardInfo->csTxDMARing[queueNumber]);

	//  Since wrap around will lock the NIC we can never use the last descriptor and 
	//  therefore we always need at least 2 descriptors.
	if (pAdapterQ->NumTxDescriptorsAvail < 2)
	{
		RtIgcLeaveCriticalLock(&pCardInfo->csTxDMARing[queueNumber]);

		if (pCardInfo->ComStats.SbXnbc != 0xffffffff)
			pCardInfo->ComStats.SbXnbc++;

		SetLastError(ERROR_DEVICE_NO_RESOURCES);
		return 0;
	}

	for (bufferIndex = 0; bufferIndex < numberOfFrames && pAdapterQ->NumTxDescriptorsAvail > 1; bufferIndex++)
	{
		ULONG descIndex = pAdapterQ->NextAvailTxDescriptor;
		volatile IGC_ADV_TRANSMIT_DESCRIPTOR* pCurrentTxDescriptor = &pAdapterQ->TxDescriptors[descIndex];
		RTNAL_FRAME* pNalFrame = pNalFrameArray[bufferIndex];

		if (pNalFrame == NULL)
		{
			errorCode = ERROR_INVALID_ADDRESS;
			break;
		}

		// Copy the NAL frame pointer to the driver array of NAL frame pointers
		pAdapterQ->pTxNalFrames[descIndex] = pNalFrame;

		// Set the physical address in the descriptor
		pCurrentTxDescriptor->read.buffer_addr = pNalFrame->frameBufferPhysicalAddr;

		// Set the frame length in the descriptor
		pCurrentTxDescriptor->read.cmd_type_len = 0xffff & pNalFrame->frameSize;

		// Set basic descriptor constants
		pCurrentTxDescriptor->read.cmd_type_len |= IGC_ADVTXD_DTYP_DATA | IGC_ADVTXD_DCMD_IFCS |
			IGC_ADVTXD_DCMD_DEXT | IGC_ADVTXD_DCMD_RS |
			IGC_ADVTXD_DCMD_EOP;
		// CurrentTxDescriptor->config.typeAndLength |= RTIGB_ADVTXD_DCMD_VLE; to enable VLAN insertion
#ifdef TIMESTAMP_SUPPORT
		if (pNalFrame->bTimeStampRequired)
			pCurrentTxDescriptor->config.typeAndLength |= IGC_ADVTXD_MAC_TSTAMP;
#endif

		//	CurrentTxDescriptor->config.paylen  = RTIGC_TXD_DMA_TXDWB;  // report the DMA time of the tx desc wb
		//ToDo
		//pCurrentTxDescriptor->config.paylen &= IXGC_ADVTXD_PAYLEN_MASK;
		pCurrentTxDescriptor->read.olinfo_status |= pNalFrame->frameSize << IGC_ADVTXD_PAYLEN_SHIFT;

		// Advance the software descriptor index
		pAdapterQ->NextAvailTxDescriptor++;
		pAdapterQ->NextAvailTxDescriptor %= pAdapterQ->NumTxDescriptors;

		// TODO: remove the commented out line when TransmitEx has been tested
		//IGC_WRITE_REG(&pCardInfo->adapter.hw, IGC_TDT(queueNumber), pAdapterQ->NextAvailTxDescriptor);
		
		//RtPrintf("***RtndTransmitEx NumTxDescriptorsAvail %d \n", pAdapterQ->NumTxDescriptorsAvail);

		// Update number of available descriptors
		pAdapterQ->NumTxDescriptorsAvail--;
	}

	if (bufferIndex > 0)
	{
		// Advance HW Tx Tail Pointer
		IGC_WRITE_REG(&pCardInfo->adapter.hw, IGC_TDT(queueNumber), pAdapterQ->NextAvailTxDescriptor);

		if (pAdapterQ->bTxMonitor)
		{
			GenerateMonitorEvent(pCardInfo->EnetAddr, (BYTE)queueNumber, RTNAL_MONITOR_EVENT_T2_TXEX, bufferIndex, 0);
		}
	}
	RtIgcLeaveCriticalLock(&pCardInfo->csTxDMARing[queueNumber]);


	// Set last error, if transmitting less frames than requested
	if (bufferIndex < numberOfFrames)
	{
		if (errorCode == 0)
			errorCode = ERROR_DEVICE_NO_RESOURCES;

		SetLastError(errorCode);
	}

	// Return number of frames passed to driver to transmit
	return bufferIndex;
}

#ifdef MULTI_QUEUE_SUPPORT
// Get destination queue, Ethernet type, and is time stamp enabled
// for a filter entry with specified index
// If the entry does not have a filter, the returned etherType is 0
//
static int GetRxFilterEntryEthertype(DRVDEV* pCardInfo,
	ULONG	filterIndex,
	ULONG* pDestinationQueue,
	USHORT* pEtherType,
	BOOL* pbAddTimeStamp,
	BOOL* pbLowLatencyInterrupt)
{
	struct ixgbe_hw* hw = pCardInfo->hw;
	volatile DWORD etqf, etqs;

	// Don't check pointer arguments since it's a static function
	// Check filter index
	if (filterIndex >= IXGBE_MAX_ETQF_FILTERS)
	{
		return ERROR_INVALID_INDEX;
	}

	// Check pointer arguments
	if (pCardInfo == NULL || pDestinationQueue == NULL || pEtherType == NULL || pbAddTimeStamp == NULL || pbLowLatencyInterrupt == NULL)
	{
		return ERROR_BAD_ARGUMENTS;
	}

	// Clear returned arguments
	*pDestinationQueue = 0;
	*pEtherType = 0;
	*pbAddTimeStamp = FALSE;
	*pbLowLatencyInterrupt = FALSE;

	// Read ETQF and ETQS registers
	etqf = IXGBE_READ_REG(hw, IXGBE_ETQF(filterIndex));
	etqs = IXGBE_READ_REG(hw, IXGBE_ETQS(filterIndex));

	// Get Ethertype and Time stamp enabled if the filter is enabled in ETQF
	if ((etqf & IXGBE_ETQF_FILTER_EN) != 0)
	{
		*pEtherType = (USHORT)(etqf & 0xffff);
		*pbAddTimeStamp = (etqf & IXGBE_ETQF_1588) != 0;
		*pbLowLatencyInterrupt = (etqs & IXGBE_ETQS_LLI) != 0;

		// Get the destination queue if queue is enabled in ETQS
		if ((etqs & IXGBE_ETQS_QUEUE_EN) != 0)
		{
			*pDestinationQueue = (etqs & IXGBE_ETQS_RX_QUEUE) >> IXGBE_ETQS_RX_QUEUE_SHIFT;
		}
	}
	return 0;
}

// Clear the receive Ethernet type filter for the specified filter entry, if it belongs to a specified queue
// Returns ERROR_NOT_OWNER, if the entry has a filter and it belongs to a another queue
//
static int ClearRxFilterEntryEthertype(DRVDEV* pCardInfo, ULONG filterIndex, ULONG destinationQueue)
{
	struct ixgbe_hw* hw = pCardInfo->hw;
	volatile DWORD etqf, etqs;
	ULONG q;

	// Don't check pointer arguments since it's a static function
	// Check filter index
	if (filterIndex >= IXGBE_MAX_ETQF_FILTERS)
	{
		return ERROR_INVALID_INDEX;
	}

	// Check valid queue
	if (destinationQueue >= (ULONG)pCardInfo->numberOfRxNicQueues)
	{
		return ERROR_BAD_ARGUMENTS;
	}

	// Read ETQF and ETQS registers
	etqf = IXGBE_READ_REG(hw, IXGBE_ETQF(filterIndex));
	etqs = IXGBE_READ_REG(hw, IXGBE_ETQS(filterIndex));

	if (etqf != 0 || etqs != 0)
	{
		q = 0;

		if ((etqs & IXGBE_ETQS_QUEUE_EN) != 0)
		{
			q = (etqs & IXGBE_ETQS_RX_QUEUE) >> IXGBE_ETQS_RX_QUEUE_SHIFT;
		}

		if (q != destinationQueue)
		{
			// The entry have a filter for another queue
			return ERROR_NOT_OWNER;
		}

		// Clear ETQF and ETQS registers
		IXGBE_WRITE_REG(hw, IXGBE_ETQF(filterIndex), 0);
		IXGBE_WRITE_REG(hw, IXGBE_ETQS(filterIndex), 0);
		IXGBE_WRITE_FLUSH(hw);
	}

	return 0;
}

// Sets the receive Ethernet type filter for the specified filter entry
// Returns error ERROR_NOT_OWNER if this filter entry is used by another queue
// Returns ERROR_ALREADY_ASSIGNED if this Ethertype is used by another filter entry
//
static int SetRxFilterEntryEthertype(DRVDEV* pCardInfo,
	ULONG	filterIndex,
	ULONG	destinationQueue,
	USHORT	etherType,
	BOOL	bAddTimeStamp,
	BOOL	bLowLatencyInterrupt)
{
	struct ixgbe_hw* hw = pCardInfo->hw;
	volatile DWORD etqf, etqs;
	USHORT etype;
	BOOL   tstamp;
	BOOL   llint;
	ULONG  q, i;
	int result;

	// Don't check pointer arguments since it's a static function
	// Check valid Ethertype and queue
	if (etherType <= RTNAL_MIN_VALID_ETHERTYPE ||
		destinationQueue >= (ULONG)pCardInfo->numberOfRxNicQueues)
	{
		return ERROR_BAD_ARGUMENTS;
	}

	// Check that this entry is not used by another queue
	// GetRxFilterEntryEthertype returns 0 and ethertype == 0, if the entry does not have a filter
	result = GetRxFilterEntryEthertype(pCardInfo, filterIndex, &q, &etype, &tstamp, &llint);
	if (result != 0)
	{
		return result;
	}

	if (etype != 0 && q != destinationQueue)
	{
		return ERROR_NOT_OWNER;
	}

	// Check that specified Ethertype is not used by anpther entry
	for (i = 0; i < IXGBE_MAX_ETQF_FILTERS; i++)
	{
		if (i != filterIndex)
		{
			GetRxFilterEntryEthertype(pCardInfo, i, &q, &etype, &tstamp, &llint);
			if (etype == etherType)
			{
				return ERROR_ALREADY_ASSIGNED;
			}
		}
	}

	// Set ETQF value
	etqf = (DWORD)etherType | IXGBE_ETQF_FILTER_EN;
	if (bAddTimeStamp)
	{
		etqf |= IXGBE_ETQF_1588;
	}

	// Set ETQS value
	etqs = (destinationQueue << IXGBE_ETQS_RX_QUEUE_SHIFT) | IXGBE_ETQS_QUEUE_EN;

	if (bLowLatencyInterrupt)
	{
		etqs |= IXGBE_ETQS_LLI;
	}

	// Write ETQF and ETQS registers
	IXGBE_WRITE_REG(hw, IXGBE_ETQF(filterIndex), etqf);
	IXGBE_WRITE_REG(hw, IXGBE_ETQS(filterIndex), etqs);
	IXGBE_WRITE_FLUSH(hw);
	return 0;
}
#endif
/**--------------------------------------------------------------------------------------------------
   \fn	RTNDPROC int RtndIoctl(void *ndp, int cmd, char *addr)

   \brief	RtndIoctl is called when the stack requires that the driver configure certain NIC
			modes or characteristics.

   \param [in,out]	ndp 	Network Device Pointer. Opaque network device identifier.
   \param	cmd				Control code to execute.
   \param [in,out]	addr	If non-null, the address.

   \return	A return value of 0 indicates Success, -1 indicates Failure.

*-----------------------------------------------------------------------------------------------**/
RTNDPROC int RtndIoctl(void* ndp, int cmd, char* addr) {

	DRVDEV* pCardInfo;
	int nStatus = 0;
	ULONG drvIndex;

	u32	fctrl_reg;

	struct igc_hw* hw;

	if (ndp == NULL)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	drvIndex = RtnGetDataLong(ndp, RTNAL_DRIVER_INDEX_OFFSET);
	if (drvIndex >= pGlobals->nCards)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	pCardInfo = pGlobals->drvCards[drvIndex];
	if (pCardInfo == NULL)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	hw = &pCardInfo->adapter.hw;

	switch (cmd)
	{
	case ENIOCPROMISC: // Turn on promiscuous mode (optional)
		RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
		Rt_igc_if_set_promisc(pCardInfo, ENIOCPROMISC);
		RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
		break;

	case ENIOCALL: // Receive good or bad. Non-promisc. (optional)
		RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
		Rt_igc_if_set_promisc(pCardInfo, ENIOCALL);
		RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
		break;

	case ENIOCNORMAL: // Receive only goood packets. (optional, default)
		RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
		fctrl_reg = IGC_READ_REG(hw, IGC_RCTL);
		fctrl_reg &= ~IGC_RCTL_SBP;
		IGC_WRITE_REG(hw, IGC_RCTL, fctrl_reg);
		RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
		break;

	case ENIOADDMULTI: // Add multicast route to this NIC (required)
	{
		//   Currrently we enable Multicast Promiscuous so that we receive all multicast packets
		//        The better way to do this is to enable just the multicast addresses that are added by this call.
		ULONG queNum = RtnGetDataLong(ndp, RTNAL_QUEUE_INDEX_OFFSET);
		if (queNum < (ULONG)pCardInfo->numberOfRxNicQueues)
		{
			RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
			// Set the flag that multicast is used by this queue
			pCardInfo->McastFlags |= 1 << queNum;

			// Set Multicast Promiscuous mode
			fctrl_reg = IGC_READ_REG(hw, IGC_RCTL);
			fctrl_reg |= IGC_RCTL_MPE;
			IGC_WRITE_REG(hw, IGC_RCTL, fctrl_reg);
			RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
		}
		else
		{
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}
		break;
	}

	case ENIODELMULTI: // Delete multicast route from this NIC (required)
		//  Since we are not tracking each multicast address we cannot delete one address
		//        addr == 0 means that the queue does not need multicast
		if (addr == 0)
		{
			ULONG queNum = RtnGetDataLong(ndp, RTNAL_QUEUE_INDEX_OFFSET);
			if (queNum < (ULONG)pCardInfo->numberOfRxNicQueues)
			{
				RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
				// Set the flag that multicast is not used by this queue
				pCardInfo->McastFlags &= ~(1 << queNum);

				// Clear Multicast Promiscuous mode if not used by any queue
				if (pCardInfo->McastFlags == 0)
				{
					fctrl_reg = IGC_READ_REG(hw, IGC_RCTL);
					fctrl_reg &= ~(IGC_RCTL_UPE | IGC_RCTL_MPE);
					IGC_WRITE_REG(hw, IGC_RCTL, fctrl_reg);
				}
				RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
			}
			else
			{
				SetLastError(ERROR_BAD_ARGUMENTS);
				nStatus = -1;
			}
		}
		break;
	case ENIOLINKSTATUS:
		if (addr == NULL)
		{
			SetLastError(ERROR_BAD_ARGUMENTS);
			return -1;
		}
		*addr = pCardInfo->LinkIsActive;
		break;

	case RTNAL_IOCTL_USE_RX_NOTIFICATIONS:
	{
		// The attached application will use receive notifications
		ULONG queNum = RtnGetDataLong(ndp, RTNAL_QUEUE_INDEX_OFFSET);
		if (queNum < (ULONG)pCardInfo->numberOfRxNicQueues)
		{
			RtIgcEnterCriticalLock(&pCardInfo->csRxDMARing[queNum]);
			pCardInfo->adapterQ[queNum].flags |= DRV_RX_NOTIFY_FLAG;
			RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[queNum]);
		}
		else
		{
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}
		break;
	}

	case RTNAL_IOCTL_RX_MONITOR:
	{
		ULONG queNum = RtnGetDataLong(ndp, RTNAL_QUEUE_INDEX_OFFSET);
		if (addr != NULL && queNum < (ULONG)pCardInfo->numberOfRxNicQueues)
		{
			pCardInfo->adapterQ[queNum].bRxMonitor = *((BOOL*)addr);
		}
		else
		{
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}
		break;
	}

	case RTNAL_IOCTL_TX_MONITOR:
	{
		ULONG queNum = RtnGetDataLong(ndp, RTNAL_QUEUE_INDEX_OFFSET);
		if (addr != NULL && queNum < (ULONG)pCardInfo->numberOfTxNicQueues)
		{
			pCardInfo->adapterQ[queNum].bTxMonitor = *((BOOL*)addr);
		}
		else
		{
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}
		break;
	}

	case RTNAL_IOCTL_SET_RX_POLLING:
	{
		// The attached application is about to start the polling timer
		ULONG queNum = RtnGetDataLong(ndp, RTNAL_QUEUE_INDEX_OFFSET);
		if (queNum < (ULONG)pCardInfo->numberOfRxNicQueues)
		{
			RtIgcEnterCriticalLock(&pCardInfo->csRxDMARing[queNum]);
			pCardInfo->adapterQ[queNum].flags |= DRV_RX_POLLING_FLAG;
			RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[queNum]);
		}
		else
		{
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}
		break;
	}

	case RTNAL_IOCTL_CLEAR_RX_POLLING:
	{
		// The attached application has stopped the polling timer
		ULONG queNum = RtnGetDataLong(ndp, RTNAL_QUEUE_INDEX_OFFSET);
		if (queNum < (ULONG)pCardInfo->numberOfRxNicQueues)
		{
			RtIgcEnterCriticalLock(&pCardInfo->csRxDMARing[queNum]);
			pCardInfo->adapterQ[queNum].flags &= ~DRV_RX_POLLING_FLAG;

			// Must flush the received packets after the polling timer has stopped
			FlushReceiveQueue(pCardInfo, queNum);
			RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[queNum]);
		}
		else
		{
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}
		break;
	}
#ifdef MULTI_QUEUE_SUPPORT
	case RTNAL_IOCTL_GET_RX_ETHERTYPE_FILTER:

		nStatus = -1;
		if (addr != NULL)
		{
			PRTNAL_ETHERTYPE_FILTER pRxFilter = (PRTNAL_ETHERTYPE_FILTER)addr;

			RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
			nStatus = GetRxFilterEntryEthertype(pCardInfo,
				pRxFilter->filterIndex,
				&pRxFilter->queueNumber,
				&pRxFilter->etherType,
				&pRxFilter->bAddTimeStamp,
				&pRxFilter->bLowLatencyInterrupt);
			if (nStatus != 0)
			{
				SetLastError(nStatus);
				nStatus = -1;
			}
			RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
		}
		else {
			SetLastError(ERROR_BAD_ARGUMENTS);
		}

		break;

	case RTNAL_IOCTL_SET_RX_ETHERTYPE_FILTER:
		nStatus = -1;
		if (addr != NULL)
		{
			PRTNAL_ETHERTYPE_FILTER pRxFilter = (PRTNAL_ETHERTYPE_FILTER)addr;
			RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
			nStatus = SetRxFilterEntryEthertype(pCardInfo,
				pRxFilter->filterIndex,
				pRxFilter->queueNumber,
				pRxFilter->etherType,
				pRxFilter->bAddTimeStamp,
				pRxFilter->bLowLatencyInterrupt);
			if (nStatus != 0)
			{
				SetLastError(nStatus);
				nStatus = -1;
			}
			RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
		}
		else {
			SetLastError(ERROR_BAD_ARGUMENTS);
		}
		break;

	case RTNAL_IOCTL_CLEAR_RX_ETHERTYPE_FILTER:
		nStatus = -1;
		if (addr != NULL)
		{
			PRTNAL_ETHERTYPE_FILTER pRxFilter = (PRTNAL_ETHERTYPE_FILTER)addr;

			RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
			nStatus = ClearRxFilterEntryEthertype(pCardInfo,
				pRxFilter->filterIndex,
				pRxFilter->queueNumber);
			if (nStatus != 0)
			{
				SetLastError(nStatus);
				nStatus = -1;
			}
			RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
		}
		else {
			SetLastError(ERROR_BAD_ARGUMENTS);
		}
		break;
#endif
#ifdef TIMESTAMP_SUPPORT
	case RTNAL_IOCTL_READ_RX_HW_TIMESTAMP:

		if (addr != NULL)
		{
			RTNAL_HW_TIME* pHwTstamps = (RTNAL_HW_TIME*)addr;

			RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
			nStatus = ReadRxHwTstamps(pCardInfo, pHwTstamps);

			if (nStatus != ERROR_SUCCESS)
			{
				SetLastError(nStatus);
				nStatus = -1;
			}

			RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
		}
		else {
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}
		break;

	case RTNAL_IOCTL_SET_RX_MESSAGE_TYPE_TO_TIMESTAMP:

		nStatus = -1;
		if (addr != NULL)
		{
			int i = 0;
			RTNAL_PTP_SET_TIMESTAMP_TYPE_ARGUMENTS* parameters = (RTNAL_PTP_SET_TIMESTAMP_TYPE_ARGUMENTS*)addr;
			enum RTNAL_PTP_MESSAGE_TYPE  messageType = parameters->messageType;
			ULONG queueNumber = parameters->queueNumber;

			nStatus = SetRxTimestampPacketType(pCardInfo, queueNumber, messageType);

			if (nStatus != 0)
			{
				SetLastError(nStatus);
				nStatus = -1;
			}
		}
		else {
			SetLastError(ERROR_BAD_ARGUMENTS);
		}
		break;

	case RTNAL_IOCTL_READ_TX_HW_TIMESTAMP:

		nStatus = -1;
		if (addr != NULL)
		{
			RTNAL_HW_TIME* pHwTstamps = (RTNAL_HW_TIME*)addr;

			RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
			nStatus = ReadTxHwTstamps(pCardInfo, pHwTstamps);

			if (nStatus != 0)
			{
				SetLastError(nStatus);
				nStatus = -1;
			}
			RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
		}
		else {
			SetLastError(ERROR_BAD_ARGUMENTS);
		}
		break;

	case RTNAL_IOCTL_SET_TX_MESSAGE_TYPE_TO_TIMESTAMP:

		nStatus = -1;
		if (addr != NULL)
		{
			RTNAL_PTP_SET_TIMESTAMP_TYPE_ARGUMENTS* parameters = (RTNAL_PTP_SET_TIMESTAMP_TYPE_ARGUMENTS*)addr;
			enum RTNAL_PTP_MESSAGE_TYPE  messageType = parameters->messageType;
			ULONG queueNumber = parameters->queueNumber;

			nStatus = SetTxTimestampPacketType(pCardInfo, queueNumber, messageType);

			if (nStatus != 0)
			{
				SetLastError(nStatus);
				nStatus = -1;
			}
		}
		else {
			SetLastError(ERROR_BAD_ARGUMENTS);
		}
		break;
	case RTNAL_IOCTL_GET_SYSTEM_TIMER:

		if (addr != NULL)
		{
			RTNAL_HW_TIME* pHwTstamps = (RTNAL_HW_TIME*)addr;
			RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
			nStatus = ReadSystemTimer(pCardInfo, pHwTstamps);

			if (nStatus != ERROR_SUCCESS)
			{
				SetLastError(nStatus);
				nStatus = -1;
			}

			RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
		}
		else {
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}

		break;
#endif
	case RTNAL_IOCTL_SET_INTERRUPT_MODERATION:

		if (addr != NULL)
		{
			int interval_ns = *((int*)addr);
			RtIgcEnterCriticalLock(&pCardInfo->csIoctl);
			nStatus = SetInterruptModeration(pCardInfo, interval_ns);

			if (nStatus != ERROR_SUCCESS)
			{
				SetLastError(nStatus);
				nStatus = -1;
			}

			RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);
		}
		else {
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}

		break;

	case RTNAL_IOCTL_READ32_REG:
		if (addr != NULL)
		{
			RTNAL_READ_WRITE_REG32* pReg32 = (RTNAL_READ_WRITE_REG32*)addr;
			pReg32->registerValue = IGC_READ_REG(hw, pReg32->registerOffset);
		}
		else {
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}
		break;

	case RTNAL_IOCTL_WRITE32_REG:

		if (addr != NULL)
		{
			RTNAL_READ_WRITE_REG32* pReg32 = (RTNAL_READ_WRITE_REG32*)addr;
			IGC_WRITE_REG(hw, pReg32->registerOffset, pReg32->registerValue);
		}
		else {
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}
		break;

	case RTNAL_IOCTL_GET_PCI_BUS_LOCATION:
		if (addr != NULL)
		{
			PRTNAL_PCI_LOCATION pLocation = (PRTNAL_PCI_LOCATION)addr;
			pLocation->busNumber = pCardInfo->BusNumber;
			pLocation->slotNumber = pCardInfo->SlotNumber;
		}
		else
		{
			SetLastError(ERROR_BAD_ARGUMENTS);
			nStatus = -1;
		}
		break;

	default:
#ifdef _DEBUG
		DebugMsg(("RtndIoctl, unsupported ioctl call.  ioctl=%x\n", cmd));
#endif
		SetLastError(ERROR_DEVICE_FEATURE_NOT_SUPPORTED);
		nStatus = -1; // unsupported ioctl call
		break;
	}

	return (nStatus);
}

/**--------------------------------------------------------------------------------------------------
   \fn	RTNDPROC VOID RtndRequest(PRTND_STATUS Status, PCHAR DeviceName, PRTND_REQUEST RtndReq)

   \brief	RtndRequest function forwards a request to the underlying driver to query the
			requested capabilities and/or statistics of the RTX64 converted NIC card.

   \param	Status	  	The RTND_STATUS type set on return from this function which reflects
						status of RtndRequest.
   \param	DeviceName	Zero terminated network adapter name.
   \param	RtndReq   	The RTND_REQUEST type specifying RTX64 network adapter capabilities and
						statistics to request.

   \return	A VOID.

   \see	RTND_REQUEST

*-----------------------------------------------------------------------------------------------**/
RTNDPROC VOID  RtndRequest(PRTND_STATUS Status, PCHAR DeviceName, PRTND_REQUEST RtndReq)
{
	ULONG i;
	DRVDEV* pCardInfo;

	int QueNum;
	struct igc_hw* hw;
	*Status = RTND_STATUS_SUCCESS;


	//Check if devicename and RtndReq has valid values
	if ((DeviceName == NULL) || (RtndReq == NULL))
	{
		DebugMsg(("RtndReq failed with bad arguments\n"));
		*Status = RTND_STATUS_FAILURE;
		return;
	}

	//Get the card of interest..
	for (i = 0; (i < pGlobals->nCards); i++)
	{
		if (!_stricmp(pGlobals->drvCards[i]->StackDeviceName, DeviceName))
		{
			break;
		}
	}
	if (i == pGlobals->nCards)
	{
		DebugMsg(("RtndReq failed. device not found\n"));
		*Status = RTND_STATUS_FAILURE;
		return;
	}
	else
		pCardInfo = pGlobals->drvCards[i];

	
	//Make sure there is enough buffer space..
	if ((RtndReq->InformationBufferLength < sizeof(ULONG)) ||
		(RtndReq->InformationBuffer == NULL))
	{
		DebugMsg(("RtndReq failed: invalid buffer length.\n"));
		*Status = RTND_STATUS_INVALID_LENGTH;
		return;
	}

	hw = &pCardInfo->adapter.hw;

	switch (RtndReq->Oid)
	{
	case RTND_OID_GEN_BYTES_RCV:
		*((DWORD*)RtndReq->InformationBuffer) = IGC_READ_REG(hw, IGC_GORCL); 
		break;
	case RTND_OID_GEN_BYTES_XMIT:
		*((DWORD*)RtndReq->InformationBuffer) = IGC_READ_REG(hw, IGC_GOTCL);
		break;

	case RTND_OID_GEN_RCV_DISCARDS:
		*((DWORD*)RtndReq->InformationBuffer) = 0;
		for (QueNum = 0; QueNum < pCardInfo->numberOfRxNicQueues; QueNum++)
		{
			*((DWORD*)RtndReq->InformationBuffer) += IGC_READ_REG(hw, IGC_RQDPC(QueNum));
		}
		break;

	case RTND_OID_GEN_XMIT_DISCARDS:
		*((DWORD*)RtndReq->InformationBuffer) += pCardInfo->ComStats.SbXnbc; //Rtndxmits
		break;

	case RTND_OID_GEN_RCV_OK:
		*((DWORD*)RtndReq->InformationBuffer) = IGC_READ_REG(hw, IGC_GPRC);
		break;

	case RTND_OID_GEN_XMIT_OK:
		*((DWORD*)RtndReq->InformationBuffer) = IGC_READ_REG(hw, IGC_GPTC);
		break;

	case RTND_OID_GEN_RCV_ERROR:
		*((DWORD*)RtndReq->InformationBuffer) = IGC_READ_REG(hw, IGC_CRCERRS) +
			IGC_READ_REG(hw, IGC_CRCERRS) +
			IGC_READ_REG(hw, IGC_ALGNERRC) +
			IGC_READ_REG(hw, IGC_RUC) +
			IGC_READ_REG(hw, IGC_ROC) +
			IGC_READ_REG(hw, IGC_RJC);
		break;

	case RTND_OID_GEN_MULTICAST_FRAMES_RCV:
		*((DWORD*)RtndReq->InformationBuffer) = IGC_READ_REG(hw, IGC_MPRC);
		break;

	case RTND_OID_GEN_MULTICAST_FRAMES_XMIT:
		*((DWORD*)RtndReq->InformationBuffer) = IGC_READ_REG(hw, IGC_MPTC);
		break;

	case RTND_OID_GEN_BROADCAST_FRAMES_RCV:
		*((DWORD*)RtndReq->InformationBuffer) = IGC_READ_REG(hw, IGC_BPRC);
		break;

	case RTND_OID_GEN_BROADCAST_FRAMES_XMIT:
		*((DWORD*)RtndReq->InformationBuffer) = IGC_READ_REG(hw, IGC_BPTC);
		break;

	case RTND_OID_GEN_STATISTICS:

		if (RtndReq->InformationBufferLength < sizeof(RTND_STATISTICS_INFO))
		{
			DebugMsg(("RtndRequest failed: RTND_OID_GEN_STATISTICS  RTND_STATUS_INVALID_LENGTH.\n"));
			*Status = RTND_STATUS_INVALID_LENGTH;
			return;
		}
		else
		{

			PRTND_STATISTICS_INFO pRtndStatistics;

			pRtndStatistics = (PRTND_STATISTICS_INFO)RtndReq->InformationBuffer;
			memset(pRtndStatistics, 0, sizeof(RTND_STATISTICS_INFO));


			pRtndStatistics->BytesRcv = IGC_READ_REG(hw, IGC_GORCL);
			pRtndStatistics->SupportedStatistics |= RTND_STATISTICS_FLAGS_VALID_BYTES_RCV;

			pRtndStatistics->BytesXmit = IGC_READ_REG(hw, IGC_GOTCL);
			pRtndStatistics->SupportedStatistics |= RTND_STATISTICS_FLAGS_VALID_BYTES_XMIT;

			pRtndStatistics->RcvDiscards = 0;
			*((DWORD*)RtndReq->InformationBuffer) = 0;
			for (QueNum = 0; QueNum < pCardInfo->numberOfRxNicQueues; QueNum++)
			{
				pRtndStatistics->RcvDiscards += IGC_READ_REG(hw, IGC_RQDPC(QueNum));
			}
			pRtndStatistics->SupportedStatistics |= RTND_STATISTICS_FLAGS_VALID_RCV_DISCARDS;

			pRtndStatistics->XmitDiscards = pCardInfo->ComStats.SbXnbc;
			pCardInfo->ComStats.SbXnbc = 0;
			pRtndStatistics->SupportedStatistics |= RTND_STATISTICS_FLAGS_VALID_XMIT_DISCARDS;

			pRtndStatistics->RcvOk = IGC_READ_REG(hw, IGC_GPRC);
			pRtndStatistics->SupportedStatistics |= RTND_STATISTICS_FLAGS_VALID_RCV_OK;

			pRtndStatistics->XmitOk = IGC_READ_REG(hw, IGC_GPTC);
			pRtndStatistics->SupportedStatistics |= RTND_STATISTICS_FLAGS_VALID_XMIT_OK;

			pRtndStatistics->RcvError = IGC_READ_REG(hw, IGC_CRCERRS) +
				IGC_READ_REG(hw, IGC_CRCERRS) +
				IGC_READ_REG(hw, IGC_ALGNERRC) +
				IGC_READ_REG(hw, IGC_RUC) +
				IGC_READ_REG(hw, IGC_ROC) +
				IGC_READ_REG(hw, IGC_RJC);
			pRtndStatistics->SupportedStatistics |= RTND_STATISTICS_FLAGS_VALID_RCV_ERROR;

			pRtndStatistics->RcvMulticastPackets = IGC_READ_REG(hw, IGC_MPRC);
			pRtndStatistics->SupportedStatistics |= RTND_STATISTICS_FLAGS_VALID_MULTICAST_BYTES_RCV;

			pRtndStatistics->XmitMulticastPackets = IGC_READ_REG(hw, IGC_MPTC);
			pRtndStatistics->SupportedStatistics |= RTND_STATISTICS_FLAGS_VALID_MULTICAST_BYTES_XMIT;

			pRtndStatistics->RcvBroadcastPackets = IGC_READ_REG(hw, IGC_BPRC);
			pRtndStatistics->SupportedStatistics |= RTND_STATISTICS_FLAGS_VALID_BROADCAST_BYTES_RCV;

			pRtndStatistics->XmitBroadcastPackets = IGC_READ_REG(hw, IGC_BPTC);
			pRtndStatistics->SupportedStatistics |= RTND_STATISTICS_FLAGS_VALID_BROADCAST_BYTES_XMIT;
		}
		break;

	case RTND_OID_GEN_MAC_ADDRESS:
		if (RtndReq->InformationBufferLength < (sizeof(BYTE) * 6))
		{
			DebugMsg(("RtndReq failed: RTND_OID_GEN_MAC_ADDRESS  RTND_STATUS_INVALID_LENGTH.\n"));
			*Status = RTND_STATUS_INVALID_LENGTH;
			return;
		}
		else
		{
			int i;
			BYTE* Enet;
			Enet = RtndReq->InformationBuffer;
			for (i = 0; i < 6; i++)
				Enet[i] = pCardInfo->EnetAddr[i];
		}
		break;

	case RTND_OID_GEN_MEDIA_CONNECT_STATUS:
		*((DWORD*)RtndReq->InformationBuffer) = pCardInfo->LinkIsActive;
		break;

	case RTND_OID_GEN_MEDIA_DUPLEX_STATE:
		*((DWORD*)RtndReq->InformationBuffer) = pCardInfo->FullDuplex;
		break;

	case RTND_OID_GEN_LINK_SPEED:
		*((DWORD*)RtndReq->InformationBuffer) = pCardInfo->LineSpeed;
		break;

	default:
		DebugMsg(("RtndReq failed: RTND_STATUS_INVALID_REQUEST. RtndReq->Oid=%x\n", RtndReq->Oid));
		*Status = RTND_STATUS_INVALID_REQUEST;
		return;
	}

	return;
}

/**--------------------------------------------------------------------------------------------------
\fn	RTNDPROC int RtndInitializeInterface(int fDisplayErrorMessages, char *DeviceName, PRTNAL_INTERFACE pNetInterface);

\brief	RtndInitializeInterface is called by NAL when an instance of a driver is being first
brought online.  This is the first routine in a driver that is called by NAL.

RtndInitializeInterface will be called once and only once per instance of the driver, but
maybe called multiple times if more than one instance of the driver is needed for
more than one NIC. As such, all drivers should keep a local configuration structure
array to hold configuration information read in during the RtndInitializeInterface function.
The driver then simply increments a local instance count variable to index the array.
RtndInitializeInterface should store a copy of the device name and a pointer to a network
device data structure in the local configuration structure array slot along with
other configuration information, so that RtndConfigure can find the proper slot in
the local configuration array.

\param	fDisplayErrorMessages	Verbose flag. If non-zero then driver has the option to be
verbose about errors or warnings. If 0, driver should be
silent and simply return the -1 error code on catastrophic
errors.
\param [in]	DeviceName   	ASCII name for this device.
Can be stored in a local configuration structure to help de-
reference network device pointer in RtndConfigure.
\param	[in,out] pNetInterface		Pointer to an RTNAL_INTERFACE defined in RtnaApi.h. This
structure contains information used by NAL to bring up
the driver. The driver updates some of information in RTNAL_INTERFACE to return back to NAL

\return	A return value of 0 indicates Success, -1 indicates Failure.
*-----------------------------------------------------------------------------------------------**/

RTNDPROC int RtndInitializeInterface(int fDisplayErrorMessages,
	char* DeviceName,
	PRTNAL_INTERFACE pNetInterface)
{
	BOOL fOk;
	DRVDEV_INIT_PARAMS InitParams;
	DRVDEV* pCardInfo;
	BOOL	fMACAddressPresent;
	
	displayErrorMessages = pGlobals->displayErrorMessages = fDisplayErrorMessages; // save 'verbose' flag
	
	if (pNetInterface->Version != RTNAL_INTERFACE_V40)
	{
		DebugMsg(("RtndInitializeInterface, Unsupported RTNAL_INTERFACE version\n"));
		return(-1);
	}

	pCardInfo = DrvCreateDevice();
	if (!pCardInfo)
	{
		return(-1);
	}

	memset(pGlobals->drvCards[pGlobals->nCards], 0, sizeof(DRVDEV));
	memset(&InitParams, 0, sizeof(DRVDEV_INIT_PARAMS));

	fMACAddressPresent = DrvGetInitParameters(pNetInterface, &InitParams);
	InitParams.pCardInfo = pGlobals->drvCards[pGlobals->nCards];

	if (InitParams.BusNumber == (PCI_MAX_BUS + 1) ||
		InitParams.DeviceNumber == (PCI_MAX_DEVICES + 1) ||
		InitParams.FuncNumber == (PCI_MAX_FUNCTION + 1))
	{
		// If PCI bus location was not provided, Search PCI/PCIe bus for network card by calling 
		// the NAL RtnEnumPciCards function. We pass a pointer to our DrvCheckForNetworkCard 
		// function which will be called back by the NAL.
		fOk = RtnEnumPciCards(DrvCheckForNetworkCard, &InitParams, fMACAddressPresent);
	}
	else
	{
		// Use PCI bus location from pNetInterface argument
		UCHAR               buffer[PCI_COMMON_HDR_LENGTH];
		PPCI_COMMON_CONFIG  PciData = (PPCI_COMMON_CONFIG)buffer;
		PCI_SLOT_NUMBER SlotNumber;
		const ULONG Offset = 0;
		ULONG bytesReturned;

		SlotNumber.u.bits.Reserved = 0;
		SlotNumber.u.bits.DeviceNumber = InitParams.DeviceNumber;
		SlotNumber.u.bits.FunctionNumber = InitParams.FuncNumber;

		bytesReturned = RtGetBusDataByOffset(
			PCIConfiguration,
			InitParams.BusNumber,
			SlotNumber.u.AsULONG,
			PciData,
			Offset,
			PCI_COMMON_HDR_LENGTH
		);

		if (bytesReturned == 0 || PciData->VendorID == PCI_INVALID_VENDORID)
		{
			DebugMsg(("No network card found at location (%u;%u;%u).\n",
				InitParams.BusNumber, InitParams.DeviceNumber, InitParams.FuncNumber));
			SetLastError(ERROR_DEVICE_ENUMERATION_ERROR);
			return (-1);
		}
		fOk = DrvCheckForNetworkCard(PciData, InitParams.BusNumber, SlotNumber, &InitParams, fMACAddressPresent);
	}

	if (!fOk)
	{
		DebugMsg(("%s can't find network card.\n", pGlobals->msgPrefix));
		return (-1);
	}

	// Initialize number of NIC queues
	strcpy(pCardInfo->StackDeviceName, DeviceName);		// we MUST store the device name for the 'config' function that follows
	DrvGetSetupParameters(pNetInterface, pCardInfo);

	// Settings to return to the caller
	pNetInterface->NumberOfRxQueues = pCardInfo->numberOfRxNicQueues;
	pNetInterface->NumberOfTxQueues = pCardInfo->numberOfTxNicQueues;
	pNetInterface->DefaultRxQueue = pCardInfo->defaultQueue;
	pNetInterface->MaxPacketSize = pCardInfo->MaxPacket; 	// TODO: should be  - ETHERNET_FCS_LEN, fix in all NIC drivers
	pNetInterface->NumRxBuffers = pCardInfo->adapterQ[0].NumRxDescriptors;
	pNetInterface->NumTxBuffers = pCardInfo->adapterQ[0].NumTxDescriptors;
	pNetInterface->LinkStatus = pCardInfo->LinkStatus;
	
	pGlobals->nCards++; // bump number of cards found counter up by 1

#ifdef _DEBUG
	DebugMsg((" built on: %s  at: %s\n", __DATE__, __TIME__));
#endif
	return (0);
}
//
// Configure MSIX multiple vectors layout
//
static BOOL
ConfigureMsixMultiVectorLayout(DRVDEV* pCardInfo)
{
	ULONG           FreeVectors;
	ULONG           ProcessorFreeVectors[MAXIMUM_PROCESSORS] = { 0 };
	ULONG           currentMessage, prn, idealProcessorForMessage;
	DWORD_PTR lowestProcessor = 0;
	DWORD_PTR highestProcessor = 0;

	FreeVectors = RtQueryProcessorVectorFreeCount(ProcessorFreeVectors);

	// Keep for debugging
	//RtPrintf("%s MessageCount %d  FreeVectors %d \n", pGlobals->msgPrefix, pCardInfo->usedMessagesCount, FreeVectors);

	// Error out if there are not enough free RTSS processor vectors
	if (pCardInfo->usedMessagesCount > FreeVectors)
	{
		DebugMsg(("MessageCount %d  > FreeVectors %d ", pGlobals->msgPrefix, pCardInfo->usedMessagesCount, FreeVectors));
		return FALSE;
	}

	// Get the ranges of realtime processors
	if (FALSE == RtGetProcessorInfo((PDWORD)&lowestProcessor, (PDWORD)&highestProcessor))
	{
		DebugMsg(("%s ConfigureMsixMultiVectorLayout failed to get realtime processors ranges.\n", pGlobals->msgPrefix));
		return (-1);
	}

	// Keep for future debugginig
	// Configure the ideal processor for other cause interrupts
	//for (prn = (int)lowestProcessor; prn <= (int)highestProcessor; prn++)
	//{
	//	DebugMsg((" %s %s FREE VECTORS CORE %d. VECT: %d \n",
	//		pGlobals->msgPrefix , pCardInfo->StackDeviceName, prn, ProcessorFreeVectors[prn]));
	//}

	// Configure the ideal processor for other cause interrupts
	for (prn = (ULONG)lowestProcessor; prn <= (ULONG)highestProcessor; prn++)
	{
		if (ProcessorFreeVectors[prn] > 0)
		{
			ProcessorFreeVectors[prn] --;
			pCardInfo->IntIdealProc = prn;
			break;
		}
	}

	// Configure the ideal processor for each Rx and Tx queue
	for (currentMessage = 0, prn = 0; currentMessage < pCardInfo->usedMessagesCount; currentMessage++)
	{
		// Is free vector found 
		BOOL found = FALSE;

		// Alternate processors for consequent queues.   
		// Start reusing once ran out.
		idealProcessorForMessage = (ULONG)lowestProcessor +
			currentMessage % ((ULONG)highestProcessor - (ULONG)lowestProcessor + 1);

		// If ideal processor has free vector , use it.
		if (ProcessorFreeVectors[idealProcessorForMessage] > 0)
		{
			prn = idealProcessorForMessage;
			found = TRUE;
		}
		else
		{
			// Ideal processor does not have free vectors, search all.
			for (prn = (ULONG)lowestProcessor; prn <= (ULONG)highestProcessor; prn++)
			{
				if (ProcessorFreeVectors[prn] > 0)
				{
					found = TRUE;
					break;
				}
			}
		}

		// There is a processor with a free vector for the current queue.
		if (TRUE == found)
		{
			// Keep for debugging
			//DebugMsg((" %s FOUND FREE VECT FOR MSG %d prn %d VECT: %d \n",
			//	pGlobals->msgPrefix, currentMessage+1, prn,
			//	ProcessorFreeVectors[prn]));

			// Claim a vector
			ProcessorFreeVectors[prn] --;
			// Assign processor to the curent message.
			// Note: Rx/Tx messages start from one hence (+1)
			pCardInfo->messageToInerruptKindQueueMap[currentMessage + 1].processor = prn;
		}
		else
		{
			DebugMsg(("%s Failed to configure MV support. Not enough free vectors.\n", pGlobals->msgPrefix));
			return FALSE;
		}

	}

	return TRUE;
}

/**--------------------------------------------------------------------------------------------------
   \fn	RTNDPROC int RtndConfigure(void *ndp)

   \brief	RtndConfigure is called for each instance of the driver when the driver should
			configure the NIC hardware. This function should set up receive buffers/transmit
			buffers, card modes, events, etc.

   \param [in,out]	ndp	Device Pointer. Opaque network device identifier.

   \return	A return value of 0 indicates Success, -1 indicates Failure.

			\note	The driver should NOT enable interrupts or packet reception on the NIC within
			RtndConfigure. The driver should wait until the RtndUpDown routine is called before
			enabling interrupts or packet reception.

*-----------------------------------------------------------------------------------------------**/
RTNDPROC int RtndConfigure(void* ndp)
{
	//  Only Setup Events, Timers and Threads
	//  Confifure the hardware in the RtndUpDown() NOT here.


	char* DevName;
	DRVDEV* pCardInfo;
	ULONG i, cidx;

	// Find the card structure which matches this device table entry
	DevName = RtnGetDeviceName(ndp);

	for (i = 0; (i < pGlobals->nCards); i++)
	{
		if (!_stricmp(pGlobals->drvCards[i]->StackDeviceName, DevName))
		{
			// Store the index for our card configuration info
			// in the device data area (index 0) for this device (ndp)
			RtnSetDataLong(ndp, RTNAL_DRIVER_INDEX_OFFSET, (unsigned long)i);
			cidx = i;
			// save the ndp (network device pointer) in our local config
			// structure to make it easy to dereference in either direction
			pGlobals->drvCards[i]->ndp = ndp;
			break;
		}
	}



	if (i == pGlobals->nCards)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		DebugMsg(("RtnGetDeviceName, invalid ndp passed to RtndConfigure().\n"));
		return (-1);
	}

	pCardInfo = pGlobals->drvCards[cidx];

	pCardInfo->DeviceIsUp = FALSE;

	if (!NicAdapterStart(pCardInfo))
	{
		DebugMsg(("RtndUpDown error.  NicAdapterStart failed\n"));
		return(-1);
	}

	// wait for the IRQ to set this correctly
	pCardInfo->LinkIsActive = RtndMediaConnectStateUnknown;

	pCardInfo->DeviceIsUp = TRUE;


	//if ((pCardInfo->LinkIsActive == RtndMediaConnectStateConnected) && (pCardInfo->hLineStatusEvent != NULL))
	//	RtSetEvent(pCardInfo->hLineStatusEvent);

	if (pCardInfo->hLineStatusEvent)
	{
		SetEvent(pCardInfo->hLineStatusEvent);
	}

	return(0);
}


/**--------------------------------------------------------------------------------------------------
\fn	RTNDPROC int RtndAttachToReceiveQueue(ULONG interfaceNum, ULONG QueueNumber, RTN_FN_PKT_CALLBACK fpCallback, void * pAppContext)

\brief	RtndAttachToReceiveQueue is called by a NAL to attach a  NAL application to the device receive queue.

Possible error codes:
ERROR_BAD_ARGUMENTS			- pAppContext is NULL or invalid interfaceNum or  QueueNumber.

\param [in]	interfaceNum	Driver instance index.
\param [in]	QueueNumber		Device queue index.
\param [in]	fpCallback		The callback to call for each receive packet.
\param [in]	pAppContext		Application context pointer.

\return	 Return the number of received packets in DMA on Success, -1 indicates Failure, use GetLastError() to get an error code.
*-----------------------------------------------------------------------------------------------**/
RTNDPROC int RtndAttachToReceiveQueue(ULONG interfaceNum, ULONG QueueNumber, RTN_FN_PKT_CALLBACK fpCallback, void* pAppContext)
{
	DRVDEV* pCardInfo;
	ULONG	i, initRxCnt = 0;
	DRV_QUE_DESCRIPTORS* pAdapterQ;

	if (pAppContext == NULL || interfaceNum >= pGlobals->nCards) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	// get pointer to the driver data structure
	pCardInfo = pGlobals->drvCards[interfaceNum];

	if (pCardInfo == NULL || QueueNumber >= (ULONG)pCardInfo->numberOfRxNicQueues) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	pAdapterQ = &pCardInfo->adapterQ[QueueNumber];

	// Lock receive DMA
	RtIgcEnterCriticalLock(&pCardInfo->csRxDMARing[QueueNumber]);

	// Walk the Rx descriptors for this Queue and add pointers
	// to application context and the callback to NAL frame headers
	for (i = 0; i < pAdapterQ->NumRxDescriptors; i++)
	{
		pAdapterQ->RxNalFrames[i].appContext = pAppContext;
		pAdapterQ->RxNalFrames[i].fpCallBack = fpCallback;
	}

	// Get the number of receive packets in DMA
	initRxCnt = GetReceiveDMACount(pCardInfo, QueueNumber);

	// Consider packets in DMA notified
	// Can't signal with RtnNotifyReceiveQueue from here,
	// because RtndAttachToReceiveQueue runs in application context
	pAdapterQ->RxDescriptorsNotified = initRxCnt;

	// The calling application is now attached
	pAdapterQ->pRxAppContext = pAppContext;
	pAdapterQ->flags |= DRV_RX_ATTACHED_FLAG;

	RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[QueueNumber]);

	//  Return the count of packets in DMA to the application.
	return initRxCnt;
}

/**--------------------------------------------------------------------------------------------------
\fn	RTNDPROC int RtndDetachReceiveQueue(ULONG interfaceNum, ULONG QueueNumber)

\brief	RtndDetachReceiveQueue is called by NAL to detach a device queue from the NAL application.

Possible error codes:
ERROR_BAD_ARGUMENTS			- Invalid interfaceNum or  QueueNumber.

\param [in]	interfaceNum	Driver instance index.
\param [in]	QueueNumber		Device queue index.

\return	A return value of 0 indicates Success, -1 indicates Failure, use GetLastError() to get an error code.
*-----------------------------------------------------------------------------------------------**/
RTNDPROC int RtndDetachReceiveQueue(ULONG interfaceNum, ULONG QueueNumber)
{
	DRVDEV* pCardInfo;
	struct igc_hw* hw;
	ULONG i;
	// filterIndex;

	if (interfaceNum >= pGlobals->nCards) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	// get pointer to the driver data structure
	pCardInfo = pGlobals->drvCards[interfaceNum];

	if (pCardInfo == NULL || QueueNumber >= (ULONG)pCardInfo->numberOfRxNicQueues) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}
	hw = &pCardInfo->adapter.hw;

	// Clear all Ethertype filters that belong to this queue
	RtIgcEnterCriticalLock(&pCardInfo->csIoctl);

	// Set the flag that multicast is not used by this queue
	pCardInfo->McastFlags &= ~(1 << QueueNumber);

	// Clear Multicast Promiscuous mode if not used by any queue
	if (pCardInfo->McastFlags == 0)
	{
		DWORD fctrl_reg = IGC_READ_REG(hw, IGC_RCTL);
		fctrl_reg &= ~IGC_RCTL_MPE;
		IGC_WRITE_REG(hw, IGC_RCTL, fctrl_reg);
	}

#ifdef MULTI_QUEUE_SUPPORT
	for (filterIndex = 0; filterIndex < IXGBE_MAX_ETQF_FILTERS; filterIndex++)
	{
		// Only clears filters for this queue (fails for another queue)
		ClearRxFilterEntryEthertype(pCardInfo, filterIndex, QueueNumber);
	}
#endif
	RtIgcLeaveCriticalLock(&pCardInfo->csIoctl);

	// Lock receive DMA
	RtIgcEnterCriticalLock(&pCardInfo->csRxDMARing[QueueNumber]);

	// Clear receive flags and application context pointer
	pCardInfo->adapterQ[QueueNumber].flags &= ~DRV_RX_FLAGS;
	pCardInfo->adapterQ[QueueNumber].pRxAppContext = NULL;

	// Clear an application context pointer and the callback pointer
	// from all NAL packet headers
	for (i = 0; i < pCardInfo->adapterQ[QueueNumber].NumRxDescriptors; i++)
	{
		pCardInfo->adapterQ[QueueNumber].RxNalFrames[i].fpCallBack = NULL;
		pCardInfo->adapterQ[QueueNumber].RxNalFrames[i].appContext = NULL;
	}

	// Discard all received packets
	FlushReceiveQueue(pCardInfo, QueueNumber);

	// Unlock receive DMA
	RtIgcLeaveCriticalLock(&pCardInfo->csRxDMARing[QueueNumber]);

	return(0);
}

/**--------------------------------------------------------------------------------------------------
\fn	RTNDPROC int RtndAttachToTransmitQueue(ULONG interfaceNum, ULONG QueueNumber);

Called by NAL application to attach a NAL application to the device transmit queue,
if the NAL application has requested transmit notifications.

\param interfaceNum	- driver interface index
\param QueueNumber	- interface queue index
\return 0 on success, -1 if ndp is invalid pointer
*-----------------------------------------------------------------------------------------------**/
RTNDPROC int
RtndAttachToTransmitQueue(ULONG interfaceNum, ULONG QueueNumber)
{
	DRVDEV* pCardInfo;
	DRV_QUE_DESCRIPTORS* pAdapterQ;
	unsigned int		i;

	if (interfaceNum >= pGlobals->nCards) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	// get pointer to the driver data structure
	pCardInfo = pGlobals->drvCards[interfaceNum];

	if (pCardInfo == NULL || QueueNumber >= (ULONG)pCardInfo->numberOfTxNicQueues)
	{
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	pAdapterQ = &pCardInfo->adapterQ[QueueNumber];

	RtIgcEnterCriticalLock(&pCardInfo->csTxDMARing[QueueNumber]);

	pAdapterQ->flags |= DRV_TX_NOTIFY;

	// Clear application-level callback for frames, transmitted by the previously attached application
	for (i = 0; i < pAdapterQ->NumTxDescriptors; i++)
	{
		if (pAdapterQ->pTxNalFrames[i] != NULL)
		{
			pAdapterQ->pTxNalFrames[i]->fpCallBack = NULL;
		}
	}

	RtIgcLeaveCriticalLock(&pCardInfo->csTxDMARing[QueueNumber]);
	return 0;
}

/**--------------------------------------------------------------------------------------------------
\fn	RTNDPROC int RtndDetachTransmitQueue(ULONG interfaceNum, ULONG QueueNumber);

Detach a driver transmit queue from the NAL queue

\param interfaceNum	- driver interface index
\param QueueNumber	- interface queue index
\return 0 on success, -1 if ndp is invalid pointer
*-----------------------------------------------------------------------------------------------**/
RTNDPROC int
RtndDetachTransmitQueue(ULONG interfaceNum, ULONG QueueNumber)
{
	DRVDEV* pCardInfo;
	DRV_QUE_DESCRIPTORS* pAdapterQ;
	unsigned int		i;

	if (interfaceNum >= pGlobals->nCards) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	// get pointer to the driver data structure
	pCardInfo = pGlobals->drvCards[interfaceNum];

	if (pCardInfo == NULL || QueueNumber >= (ULONG)pCardInfo->numberOfTxNicQueues) {
		SetLastError(ERROR_BAD_ARGUMENTS);
		return -1;
	}

	pAdapterQ = &pCardInfo->adapterQ[QueueNumber];

	RtIgcEnterCriticalLock(&pCardInfo->csTxDMARing[QueueNumber]);

	pAdapterQ->flags &= ~DRV_TX_FLAGS;

	// Clear application-level callback for frames, transmitted by the previously attached application
	for (i = 0; i < pAdapterQ->NumTxDescriptors; i++)
	{
		if (pAdapterQ->pTxNalFrames[i] != NULL)
		{
			pAdapterQ->pTxNalFrames[i]->fpCallBack = NULL;
		}
	}

	RtIgcLeaveCriticalLock(&pCardInfo->csTxDMARing[QueueNumber]);

	// Because in Common IST DRV_NOTIFY_TX_FLAG is checked without a lock,
	// wake up driver TxComplete thread 
	if (pCardInfo->hTxCompleteEvent != NULL)
	{
		SetEvent(pCardInfo->hTxCompleteEvent);
	}

	return 0;
}